@@ -11,7 +11,7 @@
 (** Lexer type *)
 
 (** Returning a stream equipped with a location function *)
-type 'te lexer_func = ?loc:Loc.t -> char Stream.t -> 'te LStream.t
+type 'te lexer_func = ?loc:Loc.t -> ?fix_loc:(Loc.t -> Loc.t) -> char Stream.t -> 'te LStream.t
 
 module type S = sig
   type te
@@ -20,7 +20,6 @@ module type S = sig
   val tok_pattern_strings : 'c pattern -> string * string option
   val tok_func : te lexer_func
   val tok_using : 'c pattern -> unit
-  val tok_removing : 'c pattern -> unit
   val tok_match : 'c pattern -> te -> 'c
   val tok_text : 'c pattern -> string
 
@@ -32,7 +31,6 @@ module type S = sig
     val get : unit -> t
     val drop : unit -> unit
     val get_comments : t -> ((int * int) * string) list
-    val set_loc_offset : int -> unit
   end
 
 end
