@@ -87,13 +87,21 @@ let ide_doc = ref None
 let get_doc () = Option.get !ide_doc
 let set_doc doc = ide_doc := Some doc
 
+(* This patches up the offset to compensate the XML protocol not
+   having a notion of absolute offset *)
+let fix_loc ~off loc =
+  Loc.{ loc with bol_pos = loc.bol_pos + off;
+                 bp = loc.bp + off;
+                 ep = loc.ep + off}
+
 let add ((((s,eid),(sid,verbose)),off),(line_nb,bol_pos)) =
   let doc = get_doc () in
   let open Loc in
   (* note: this won't yield correct values for bol_pos_last,
      but the debugger doesn't use that *)
   let loc = { (initial ToplevelInput) with bp=off; line_nb } in
-  let pa = Pcoq.Parsable.make ~loc (Stream.of_string s) in
+  let fix_loc = fix_loc ~off in
+  let pa = Pcoq.Parsable.make ~loc ~fix_loc (Gramlib.Stream.of_string s) in
   match Stm.parse_sentence ~doc sid ~entry:Pvernac.main_entry pa with
   | None -> assert false (* s may not be empty *)
   | Some ast ->
@@ -120,13 +128,13 @@ let edit_at id =
  * be removed in the next version of the protocol.
  *)
 let query (route, (s,id)) =
-  let pa = Pcoq.Parsable.make (Stream.of_string s) in
+  let pa = Pcoq.Parsable.make (Gramlib.Stream.of_string s) in
   let doc = get_doc () in
   Stm.query ~at:id ~doc ~route pa
 
 let annotate phrase =
   let doc = get_doc () in
-  let pa = Pcoq.Parsable.make (Stream.of_string phrase) in
+  let pa = Pcoq.Parsable.make (Gramlib.Stream.of_string phrase) in
   match Stm.parse_sentence ~doc (Stm.get_current_state ~doc) ~entry:Pvernac.main_entry pa with
   | None -> Richpp.richpp_of_pp ~width:78 (Pp.mt ())
   | Some ast ->
@@ -206,14 +214,14 @@ let process_goal short sigma g =
   in
   { Interface.goal_hyp = List.rev hyps; Interface.goal_ccl = ccl; Interface.goal_id = Proof.goal_uid g; Interface.goal_name = name }
 
-let process_goal_diffs diff_goal_map oldp nsigma ng =
+let process_goal_diffs ~short diff_goal_map oldp nsigma ng =
   let env = Global.env () in
   let name = if Printer.print_goal_names () then Some (Names.Id.to_string (Termops.evar_suggested_name env nsigma ng)) else None in
   let og_s = match oldp, diff_goal_map with
   | Some oldp, Some diff_goal_map -> Proof_diffs.map_goal ng diff_goal_map
   | None, _ | _, None -> None
   in
-  let (hyps_pp_list, concl_pp) = Proof_diffs.diff_goal ?og_s (Proof_diffs.make_goal env nsigma ng) in
+  let (hyps_pp_list, concl_pp) = Proof_diffs.diff_goal ~short ?og_s (Proof_diffs.make_goal env nsigma ng) in
   { Interface.goal_hyp = hyps_pp_list; Interface.goal_ccl = concl_pp;
     Interface.goal_id = Proof.goal_uid ng; Interface.goal_name = name }
 
@@ -237,7 +245,7 @@ let export_pre_goals flags Proof.{ sigma; goals; stack } process =
 
 let subgoals flags =
   let doc = get_doc () in
-  set_doc @@ Stm.finish ~doc;
+  set_doc @@ fst @@ Stm.finish ~doc;
   let short = match flags.Interface.gf_mode with
   | "short" -> true
   | _ -> false
@@ -251,7 +259,7 @@ let subgoals flags =
         | None -> None
         | Some oldp -> Some (Proof_diffs.make_goal_map oldp newp)
         in
-        Some (export_pre_goals flags Proof.(data newp) (process_goal_diffs diff_goal_map oldp))
+        Some (export_pre_goals flags Proof.(data newp) (process_goal_diffs ~short diff_goal_map oldp))
        with Pp_diff.Diff_Failure msg ->
          Proof_diffs.notify_proof_diff_failure msg;
          Some (export_pre_goals flags Proof.(data newp) (process_goal short)))
@@ -268,7 +276,7 @@ let goals () =
 let evars () =
   try
     let doc = get_doc () in
-    set_doc @@ Stm.finish ~doc;
+    set_doc @@ fst @@ Stm.finish ~doc;
     let pfts = Vernacstate.Declare.give_me_the_proof () in
     let Proof.{ sigma } = Proof.data pfts in
     let exl = Evar.Map.bindings (Evd.undefined_map sigma) in
@@ -302,9 +310,9 @@ let status force =
   (* We remove the initial part of the current [DirPath.t]
      (usually Top in an interactive session, cf "coqtop -top"),
      and display the other parts (opened sections and modules) *)
-  set_doc (Stm.finish ~doc:(get_doc ()));
+  set_doc (fst @@ Stm.finish ~doc:(get_doc ()));
   if force then
-    set_doc (Stm.join ~doc:(get_doc ()));
+    set_doc (fst @@ Stm.join ~doc:(get_doc ()));
   let path =
     let l = Names.DirPath.repr (Lib.cwd ()) in
     List.rev_map Names.Id.to_string l
@@ -478,7 +486,7 @@ let init =
            get_doc (), init_sid, `NewTip in
          if Filename.check_suffix file ".v" then
            Stm.set_compilation_hints file;
-         set_doc (Stm.finish ~doc);
+         set_doc (fst @@ Stm.finish ~doc);
          initial_id
    end
 
@@ -574,7 +582,7 @@ let msg_format = ref (fun () ->
 let loop ( { Coqtop.run_mode; color_mode },_) ~opts:_ state =
   match run_mode with
   | Coqtop.Batch -> exit 0
-  | Coqtop.(Query PrintTags) -> Coqtop.print_style_tags color_mode; exit 0
+  | Coqtop.(Query PrintTags) -> Colors.print_style_tags color_mode; exit 0
   | Coqtop.(Query _) -> Printf.eprintf "Unknown query"; exit 1
   | Coqtop.Interactive ->
   let open Vernac.State in
