@@ -33,40 +33,39 @@ open Util
    union-find algorithm. The assertions $<$ and $\le$ are represented by
    adjacency lists *)
 
-module RawLevel =
-struct
+module UGlobal = struct
   open Names
 
-  module UGlobal = struct
-    type t = {
-      library : DirPath.t;
-      process : string;
-      uid : int;
-    }
+  type t = {
+    library : DirPath.t;
+    process : string;
+    uid : int;
+  }
 
-    let make library process uid = { library; process; uid }
+  let make library process uid = { library; process; uid }
 
-    let repr x = (x.library, x.process, x.uid)
+  let repr x = (x.library, x.process, x.uid)
 
-    let equal u1 u2 =
-      Int.equal u1.uid u2.uid &&
-      DirPath.equal u1.library u2.library &&
-      String.equal u1.process u2.process
+  let equal u1 u2 =
+    Int.equal u1.uid u2.uid &&
+    DirPath.equal u1.library u2.library &&
+    String.equal u1.process u2.process
 
-    let hash u = Hashset.Combine.combine3 u.uid (String.hash u.process) (DirPath.hash u.library)
+  let hash u = Hashset.Combine.combine3 u.uid (String.hash u.process) (DirPath.hash u.library)
 
-    let compare u1 u2 =
-      let c = Int.compare u1.uid u2.uid in
+  let compare u1 u2 =
+    let c = Int.compare u1.uid u2.uid in
+    if c <> 0 then c
+    else
+      let c = DirPath.compare u1.library u2.library in
       if c <> 0 then c
-      else
-        let c = DirPath.compare u1.library u2.library in
-        if c <> 0 then c
-        else String.compare u1.process u2.process
-  end
+      else String.compare u1.process u2.process
+end
+
+module RawLevel =
+struct
 
   type t =
-    | SProp
-    | Prop
     | Set
     | Level of UGlobal.t
     | Var of int
@@ -76,8 +75,6 @@ struct
   let equal x y =
     x == y ||
       match x, y with
-      | SProp, SProp -> true
-      | Prop, Prop -> true
       | Set, Set -> true
       | Level l, Level l' -> UGlobal.equal l l'
       | Var n, Var n' -> Int.equal n n'
@@ -85,12 +82,6 @@ struct
 
   let compare u v =
     match u, v with
-    | SProp, SProp -> 0
-    | SProp, _ -> -1
-    | _, SProp -> 1
-    | Prop,Prop -> 0
-    | Prop, _ -> -1
-    | _, Prop -> 1
     | Set, Set -> 0
     | Set, _ -> -1
     | _, Set -> 1
@@ -102,8 +93,6 @@ struct
   let hequal x y =
     x == y ||
       match x, y with
-      | SProp, SProp -> true
-      | Prop, Prop -> true
       | Set, Set -> true
       | UGlobal.(Level { library = d; process = s; uid = n }, Level  { library = d'; process = s'; uid = n' }) ->
         n == n' && s==s' && d == d'
@@ -111,8 +100,6 @@ struct
       | _ -> false
 
   let hcons = function
-    | SProp as x -> x
-    | Prop as x -> x
     | Set as x -> x
     | UGlobal.(Level { library = d; process = s; uid = n }) as x ->
       let s' = CString.hcons s in
@@ -123,8 +110,6 @@ struct
   open Hashset.Combine
 
   let hash = function
-    | SProp -> combinesmall 1 0
-    | Prop -> combinesmall 1 1
     | Set -> combinesmall 1 2
     | Var n -> combinesmall 2 n
     | Level l -> combinesmall 3 (UGlobal.hash l)
@@ -133,11 +118,7 @@ end
 
 module Level = struct
 
-  module UGlobal = RawLevel.UGlobal
-
   type raw_level = RawLevel.t =
-  | SProp
-  | Prop
   | Set
   | Level of UGlobal.t
   | Var of int
@@ -173,40 +154,24 @@ module Level = struct
   let make l = hcons { hash = RawLevel.hash l; data = l }
 
   let set = make Set
-  let prop = make Prop
-  let sprop = make SProp
 
   let is_small x =
     match data x with
     | Level _ -> false
     | Var _ -> false
-    | SProp -> true
-    | Prop -> true
     | Set -> true
 
-  let is_prop x =
-    match data x with
-    | Prop -> true
-    | _ -> false
-
   let is_set x =
     match data x with
     | Set -> true
     | _ -> false
 
-  let is_sprop x =
-    match data x with
-    | SProp -> true
-    | _ -> false
-
   let compare u v =
     if u == v then 0
     else RawLevel.compare (data u) (data v)
 
   let to_string x =
     match data x with
-    | SProp -> "SProp"
-    | Prop -> "Prop"
     | Set -> "Set"
     | UGlobal.(Level { library = d; process = s; uid = n }) ->
       Names.DirPath.to_string d ^
@@ -275,16 +240,8 @@ module Level = struct
 
 end
 
-(* Deprecated *)
-module LMap = Level.Map
-module LSet = Level.Set
-
-type 'a universe_map = 'a Level.Map.t
-
 type universe_level = Level.t
 
-type universe_level_subst_fn = universe_level -> universe_level
-
 type universe_set = Level.Set.t
 
 (* An algebraic universe [universe] is either a universe variable
@@ -339,8 +296,6 @@ struct
         if Int.equal 0 c then  Level.compare x x'
         else c
 
-    let sprop = hcons (Level.sprop, 0)
-    let prop = hcons (Level.prop, 0)
     let set = hcons (Level.set, 0)
     let type1 = hcons (Level.set, 1)
 
@@ -354,23 +309,10 @@ struct
 
     let hash = ExprHash.hash
 
-    let leq (u,n) (v,n') =
-      let cmp = Level.compare u v in
-        if Int.equal cmp 0 then n <= n'
-        else if n <= n' then
-          (Level.is_prop u && not (Level.is_sprop v))
-        else false
-
     let successor (u,n as e) =
       if is_small e then type1
       else (u, n + 1)
 
-    let addn k (u,n as x) =
-      if k = 0 then x
-      else if Level.is_small u then
-        (Level.set,n+k)
-      else (u,n+k)
-
     type super_result =
         SuperSame of bool
         (* The level expressions are in cumulativity relation. boolean
@@ -386,16 +328,7 @@ struct
     let super (u,n) (v,n') =
       let cmp = Level.compare u v in
         if Int.equal cmp 0 then SuperSame (n < n')
-        else
-          let open RawLevel in
-          match Level.data u, n, Level.data v, n' with
-          | SProp, _, SProp, _ | Prop, _, Prop, _ -> SuperSame (n < n')
-          | SProp, 0, Prop, 0 -> SuperSame true
-          | Prop, 0, SProp, 0 -> SuperSame false
-          | (SProp | Prop), 0, _, _ -> SuperSame true
-          | _, _, (SProp | Prop), 0 -> SuperSame false
-
-          | _, _, _, _ -> SuperDiff cmp
+        else SuperDiff cmp
 
     let to_string (v, n) =
       if Int.equal n 0 then Level.to_string v
@@ -420,8 +353,6 @@ struct
     let map f (v, n as x) =
       let v' = f v in
         if v' == v then x
-        else if Level.is_prop v' && n != 0 then
-          (Level.set, n)
         else (v', n)
 
   end
@@ -441,7 +372,7 @@ struct
 
   module Huniv = Hashcons.Hlist(Expr)
 
-  let hcons = Hashcons.recursive_hcons Huniv.generate Huniv.hcons Expr.hcons
+  let hcons = Hashcons.simple_hcons Huniv.generate Huniv.hcons Expr.hcons
 
   module Self = struct type nonrec t = t let compare = compare end
   module Map = CMap.Make(Self)
@@ -477,19 +408,17 @@ struct
     | _ -> None
 
   let levels l =
-    List.fold_left (fun acc x -> Level.Set.add (Expr.get_level x) acc) Level.Set.empty l
+    let fold acc x =
+      let l = Expr.get_level x in
+      Level.Set.add l acc
+    in
+    List.fold_left fold Level.Set.empty l
 
   let is_small u =
     match u with
     | [l] -> Expr.is_small l
     | _ -> false
 
-  let sprop = tip Expr.sprop
-
-  (* The lower predicative level of the hierarchy that contains (impredicative)
-     Prop and singleton inductive types *)
-  let type0m = tip Expr.prop
-
   (* The level of sets *)
   let type0 = tip Expr.set
 
@@ -497,8 +426,6 @@ struct
      hence the definition of [type1_univ], the type of [Prop] *)
   let type1 = tip Expr.type1
 
-  let is_sprop x = equal sprop x
-  let is_type0m x = equal type0m x
   let is_type0 x = equal type0 x
 
   (* Returns the formal universe that lies just above the universe variable u.
@@ -508,9 +435,6 @@ struct
     else
       List.Smart.map (fun x -> Expr.successor x) l
 
-  let addn n l =
-    List.Smart.map (fun x -> Expr.addn n x) l
-
   let rec merge_univs l1 l2 =
     match l1, l2 with
     | [], _ -> l2
@@ -550,28 +474,8 @@ struct
   let repr x : t = x
 end
 
-type universe = Universe.t
-
-(* The level of predicative Set *)
-let type0m_univ = Universe.type0m
-let type0_univ = Universe.type0
-let type1_univ = Universe.type1
-let is_type0m_univ = Universe.is_type0m
-let is_type0_univ = Universe.is_type0
-let is_univ_variable l = Universe.level l != None
-let is_small_univ = Universe.is_small
-let pr_uni = Universe.pr
-
-let sup = Universe.sup
-let super = Universe.super
-
-let universe_level = Universe.level
-
-
 type constraint_type = AcyclicGraph.constraint_type = Lt | Le | Eq
 
-type explanation = (constraint_type * Level.t) list
-
 let constraint_type_ord c1 c2 = match c1, c2 with
 | Lt, Lt -> 0
 | Lt, _ -> -1
@@ -616,12 +520,6 @@ struct
 
 end
 
-module Constraint = Constraints
-
-let empty_constraint = Constraints.empty
-let union_constraint = Constraints.union
-let eq_constraint = Constraints.equal
-
 type constraints = Constraints.t
 
 module Hconstraint =
@@ -667,54 +565,6 @@ let enforce_eq_level u v c =
   if Level.equal u v then c
   else Constraints.add (u,Eq,v) c
 
-let enforce_eq u v c =
-  match Universe.level u, Universe.level v with
-    | Some u, Some v -> enforce_eq_level u v c
-    | _ -> anomaly (Pp.str "A universe comparison can only happen between variables.")
-
-let check_univ_eq u v = Universe.equal u v
-
-let enforce_eq u v c =
-  if check_univ_eq u v then c
-  else enforce_eq u v c
-
-let constraint_add_leq v u c =
-  (* We just discard trivial constraints like u<=u *)
-  if Universe.Expr.equal v u then c
-  else
-    match v, u with
-    | (x,n), (y,m) ->
-    let j = m - n in
-      if j = -1 (* n = m+1, v+1 <= u <-> v < u *) then
-        Constraints.add (x,Lt,y) c
-      else if j <= -1 (* n = m+k, v+k <= u and k>0 *) then
-        if Level.equal x y then (* u+k <= u with k>0 *)
-          Constraints.add (x,Lt,x) c
-        else anomaly (Pp.str"Unable to handle arbitrary u+k <= v constraints.")
-      else if j = 0 then
-        Constraints.add (x,Le,y) c
-      else (* j >= 1 *) (* m = n + k, u <= v+k *)
-        if Level.equal x y then c (* u <= u+k, trivial *)
-        else if Level.is_small x then c (* Prop,Set <= u+S k, trivial *)
-        else Constraints.add (x,Le,y) c (* u <= v implies u <= v+k *)
-
-let check_univ_leq_one u v = Universe.exists (Universe.Expr.leq u) v
-
-let check_univ_leq u v =
-  Universe.for_all (fun u -> check_univ_leq_one u v) u
-
-let enforce_leq u v c =
-  match Universe.is_sprop u, Universe.is_sprop v with
-  | true, true -> c
-  | true, false -> Constraints.add (Level.sprop,Le,Level.prop) c
-  | false, true -> Constraints.add (Level.prop,Le,Level.sprop) c
-  | false, false ->
-    List.fold_left (fun c v -> (List.fold_left (fun c u -> constraint_add_leq u v c) c u)) c v
-
-let enforce_leq u v c =
-  if check_univ_leq u v then c
-  else enforce_leq u v c
-
 let enforce_leq_level u v c =
   if Level.equal u v then c else Constraints.add (u,Le,v) c
 
@@ -739,10 +589,7 @@ let univ_level_rem u v min =
 (** A universe level substitution, note that no algebraic universes are
     involved *)
 
-type universe_level_subst = universe_level universe_map
-
-(** A full substitution might involve algebraic universes *)
-type universe_subst = universe universe_map
+type universe_level_subst = universe_level Level.Map.t
 
 module Variance =
 struct
@@ -814,7 +661,7 @@ module Instance : sig
 
     val share : t -> t * int
 
-    val subst_fn : universe_level_subst_fn -> t -> t
+    val subst_fn : (Level.t -> Level.t) -> t -> t
 
     val pr : (Level.t -> Pp.t) -> ?variance:Variance.t array -> t -> Pp.t
     val levels : t -> Level.Set.t
@@ -879,7 +726,6 @@ struct
     else Array.append x y
 
   let of_array a =
-    assert(Array.for_all (fun x -> not (Level.is_prop x || Level.is_sprop x)) a);
     a
 
   let to_array a = a
@@ -1019,8 +865,6 @@ struct
 
 end
 
-module AUContext = AbstractContext (* Deprecated *)
-
 type 'a univ_abstracted = {
   univ_abstracted_value : 'a;
   univ_abstracted_binder : AbstractContext.t;
@@ -1141,42 +985,12 @@ let subst_univs_level_constraints subst csts =
 let subst_univs_level_abstract_universe_context subst (inst, csts) =
   inst, subst_univs_level_constraints subst csts
 
-(** With level to universe substitutions. *)
-type universe_subst_fn = universe_level -> universe
-
-let make_subst subst = fun l -> Level.Map.find l subst
-
-let subst_univs_expr_opt fn (l,n) =
-  Universe.addn n (fn l)
-
-let subst_univs_universe fn ul =
-  let subst, nosubst =
-    List.fold_right (fun u (subst,nosubst) ->
-      try let a' = subst_univs_expr_opt fn u in
-            (a' :: subst, nosubst)
-      with Not_found -> (subst, u :: nosubst))
-      ul ([], [])
-  in
-    if CList.is_empty subst then ul
-    else
-      let substs =
-        List.fold_left Universe.merge_univs [] subst
-      in
-        List.fold_left (fun acc u -> Universe.merge_univs acc (Universe.tip u))
-          substs nosubst
-
 let make_instance_subst i =
   let arr = Instance.to_array i in
     Array.fold_left_i (fun i acc l ->
       Level.Map.add l (Level.var i) acc)
       Level.Map.empty arr
 
-let make_inverse_instance_subst i =
-  let arr = Instance.to_array i in
-    Array.fold_left_i (fun i acc l ->
-      Level.Map.add (Level.var i) l acc)
-      Level.Map.empty arr
-
 let make_abstract_instance (ctx, _) =
   Array.init (Array.length ctx) (fun i -> Level.var i)
 
@@ -1215,9 +1029,6 @@ let pr_abstract_universe_context = AbstractContext.pr
 
 let pr_universe_context_set = ContextSet.pr
 
-let pr_universe_subst =
-  Level.Map.pr (fun u -> str" := " ++ Universe.pr u ++ spc ())
-
 let pr_universe_level_subst =
   Level.Map.pr (fun u -> str" := " ++ Level.pr u ++ spc ())
 
@@ -1241,30 +1052,4 @@ let hcons_universe_context_set (v, c) =
 
 let hcons_univ x = Universe.hcons x
 
-(* Universe inconsistency: error raised when trying to enforce a relation
-   that would create a cycle in the graph of universes. *)
-
-type univ_inconsistency = constraint_type * universe * universe * explanation Lazy.t option
-
-(* Do not use in this file as we may be type-in-type *)
-exception UniverseInconsistency of univ_inconsistency
-
-let explain_universe_inconsistency prl (o,u,v,p : univ_inconsistency) =
-  let pr_uni = Universe.pr_with prl in
-  let pr_rel = function
-    | Eq -> str"=" | Lt -> str"<" | Le -> str"<="
-  in
-  let reason = match p with
-    | None -> mt()
-    | Some p ->
-      let p = Lazy.force p in
-      if p = [] then mt ()
-      else
-        str " because" ++ spc() ++ pr_uni v ++
-        prlist (fun (r,v) -> spc() ++ pr_rel r ++ str" " ++ prl v)
-          p ++
-        (if Universe.equal (Universe.make (snd (List.last p))) u then mt() else
-           (spc() ++ str "= " ++ pr_uni u))
-  in
-    str "Cannot enforce" ++ spc() ++ pr_uni u ++ spc() ++
-      pr_rel o ++ spc() ++ pr_uni v ++ reason
+type explanation = Level.t * (constraint_type * Level.t) list
