@@ -17,6 +17,7 @@ open Names
 open Vmvalues
 open Vmbytecodes
 open Vmemitcodes
+open Genlambda
 open Vmlambda
 open Constr
 open Declarations
@@ -589,7 +590,7 @@ let rec compile_lam env cenv lam sz cont =
     | _ -> comp_app (compile_lam env) (compile_lam env) cenv f args sz cont
     end
 
-  | Lfix ((rec_args, init), (_decl, types, bodies)) ->
+  | Lfix ((rec_args, _, init), (_decl, types, bodies)) ->
       let ndef = Array.length types in
       let rfv = ref empty_fv in
       let lbl_types = Array.make ndef Label.no in
@@ -657,7 +658,7 @@ let rec compile_lam env cenv lam sz cont =
       compile_fv cenv fv.fv_rev sz
         (Kclosurecofix(fv.size, init, lbl_types, lbl_bodies) :: cont)
 
-  | Lcase(ci,rtbl,t,a,branches) ->
+  | Lcase ((ci, rtbl, _), t, a, branches) ->
       let ind = ci.ci_ind in
       let mib = lookup_mind (fst ind) env in
       let oib = mib.mind_packets.(snd ind) in
@@ -752,12 +753,21 @@ let rec compile_lam env cenv lam sz cont =
       in
       compile_lam env cenv a sz code_sw
 
-  | Lmakeblock (tag,args) ->
+  | Lmakeblock (_, tag, args) ->
     let arity = Array.length args in
     let cont = code_makeblock ~stack_size:(sz+arity-1) ~arity ~tag cont in
     if Int.equal arity 0 then cont
     else comp_args (compile_lam env) cenv args sz cont
 
+  | Lparray (args, def) ->
+    (* Hack: brutally pierce through the abstraction of PArray *)
+    let dummy = KerName.make (ModPath.MPfile DirPath.empty) (Names.Label.of_id @@ Id.of_string "dummy") in
+    let dummy = (MutInd.make1 dummy, 0) in
+    let ar = Lmakeblock (dummy, 0, args) in (* build the ocaml array *)
+    let kind = Lmakeblock (dummy, 0, [|ar; def|]) in (* Parray.Array *)
+    let v = Lmakeblock (dummy, 0, [|kind|]) (* the reference *) in
+    compile_lam env cenv v sz cont
+
   | Lprim (kn, op, args) ->
 
     begin match get_caml_prim op with
@@ -784,6 +794,8 @@ let rec compile_lam env cenv lam sz cont =
       comp_args (compile_lam env) cenv args sz (Kprim(op, kn)::cont)
     end
 
+  | Lforce -> CErrors.anomaly Pp.(str "The VM should not use force")
+
 and compile_get_global cenv (kn,u) sz cont =
   set_max_stack_size sz;
   if Univ.Instance.is_empty u then
@@ -899,8 +911,7 @@ let compile_constant_body ~fail_on_error env univs = function
             let con= Constant.make1 (Constant.canonical kn') in
               Some (BCalias (get_alias env con))
         | _ ->
-            let sigma _ = assert false in
-            let res = compile ~fail_on_error ~universes:instance_size env sigma body in
+            let res = compile ~fail_on_error ~universes:instance_size env empty_evars body in
             Option.map (fun (code, fv) -> BCdefined (code, fv)) res
 
 (* Shortcut of the previous function used during module strengthening *)
