@@ -98,7 +98,7 @@ module EntryDataMap = EntryCommand.Map(EntryData)
 
 type ext_kind =
   | ByGrammar of extend_rule
-  | ByEXTEND of (unit -> unit) * (unit -> unit)
+  | ByEXTEND of string * (unit -> unit) * (unit -> unit)
   | ByEntry : ('a * 'b) EntryCommand.tag * string * 'b Entry.t -> ext_kind
 
 (** The list of extensions *)
@@ -134,7 +134,7 @@ let grammar_delete_reinit e reinit d =
 let grammar_extend e ext =
   let undo () = grammar_delete e ext in
   let redo () = safe_extend e ext in
-  camlp5_state := ByEXTEND (undo, redo) :: !camlp5_state;
+  camlp5_state := ByEXTEND (Entry.name e, undo, redo) :: !camlp5_state;
   redo ()
 
 let grammar_extend_sync e ext =
@@ -159,15 +159,18 @@ let rec remove_grammars n =
            camlp5_state := t;
            remove_grammars (n-1)
        | ByGrammar (ExtendRule (g, ext)) :: t ->
-           grammar_delete g ext;
-           camlp5_state := t;
-           remove_grammars (n-1)
-       | ByEXTEND (undo,redo)::t ->
+         (try grammar_delete g ext
+          with Not_found -> Feedback.msg_info Pp.(str "failed on " ++ str (Entry.name g));
+            raise Not_found);
+         camlp5_state := t;
+         remove_grammars (n-1)
+       | ByEXTEND (name, undo,redo)::t ->
+         Feedback.msg_info Pp.(str "backtracking and redoing byextend on " ++ str name);
            undo();
            camlp5_state := t;
            remove_grammars n;
            redo();
-           camlp5_state := ByEXTEND (undo,redo) :: !camlp5_state
+           camlp5_state := ByEXTEND (name, undo,redo) :: !camlp5_state
        | ByEntry (tag, name, e) :: t ->
            Unsafe.clear_entry e;
            camlp5_state := t;
@@ -199,28 +202,6 @@ let parse_string f ?loc x =
   let strm = Stream.of_string x in
   Entry.parse f (Parsable.make ?loc strm)
 
-(* universes not used by Coq build but still used by some plugins *)
-type gram_universe = string
-
-let utables : (string, unit) Hashtbl.t =
-  Hashtbl.create 97
-
-let create_universe u =
-  let () = Hashtbl.add utables u () in
-  u
-
-let uprim   = create_universe "prim" [@@deprecated "Deprecated in 8.13"]
-let uconstr = create_universe "constr" [@@deprecated "Deprecated in 8.13"]
-let utactic = create_universe "tactic" [@@deprecated "Deprecated in 8.13"]
-
-let get_univ u =
-  if Hashtbl.mem utables u then u
-  else raise Not_found
-
-let new_entry _ s =
-  let e = Entry.make s in
-  e
-
 module GrammarObj =
 struct
   type ('r, _, _) obj = 'r Entry.t
@@ -252,11 +233,7 @@ let create_generic_entry2 (type a) s (etyp : a raw_abstract_argument_type) : a E
   let () = Grammar.register0 t e in
   e
 
-let create_generic_entry (type a) _ s (etyp : a raw_abstract_argument_type) : a Entry.t =
-  create_generic_entry2 s etyp
-
 (* Initial grammar entries *)
-
 module Prim =
   struct
 
@@ -271,6 +248,7 @@ module Prim =
     let string = Entry.create "string"
     let lstring = Entry.create "lstring"
     let reference = Entry.create "reference"
+    let fields = Entry.create "fields"
     let by_notation = Entry.create "by_notation"
     let smart_global = Entry.create "smart_global"
     let strategy_level = Entry.create "strategy_level"
@@ -284,7 +262,6 @@ module Prim =
     let univ_decl = Entry.create "univ_decl"
     let ident_decl = Entry.create "ident_decl"
     let pattern_ident = Entry.create "pattern_ident"
-    let pattern_identref = pattern_ident (* To remove in 8.14 *)
 
     (* A synonym of ident - maybe ident will be located one day *)
     let base_ident = Entry.create "base_ident"
@@ -306,7 +283,6 @@ module Constr =
     (* Entries that can be referred via the string -> Entry.t table *)
     let constr = Entry.create "constr"
     let term = Entry.create "term"
-    let operconstr = term
     let constr_eoi = eoi_entry constr
     let lconstr = Entry.create "lconstr"
     let binder_constr = Entry.create "binder_constr"
@@ -319,7 +295,6 @@ module Constr =
     let pattern = Entry.create "pattern"
     let constr_pattern = Entry.create "constr_pattern"
     let cpattern = Entry.create "cpattern"
-    let lconstr_pattern = cpattern
     let closed_binder = Entry.create "closed_binder"
     let binder = Entry.create "binder"
     let binders = Entry.create "binders"
@@ -330,7 +305,6 @@ module Constr =
     let typeclass_constraint = Entry.create "typeclass_constraint"
     let record_declaration = Entry.create "record_declaration"
     let arg = Entry.create "arg"
-    let appl_arg = arg
     let type_cstr = Entry.create "type_cstr"
   end
 
@@ -351,7 +325,10 @@ let epsilon_value (type s tr a) f (e : (s, tr, a) Symbol.t) =
 
 module GramState = Store.Make ()
 
-type 'a grammar_extension = 'a -> GramState.t -> extend_rule list * GramState.t
+type 'a grammar_extension = {
+  gext_fun : 'a -> GramState.t -> extend_rule list * GramState.t;
+  gext_eq : 'a -> 'a -> bool;
+}
 
 module GrammarCommand = Dyn.Make ()
 module GrammarInterp = struct type 'a t = 'a grammar_extension end
@@ -359,7 +336,10 @@ module GrammarInterpMap = GrammarCommand.Map(GrammarInterp)
 
 let grammar_interp = ref GrammarInterpMap.empty
 
-type ('a, 'b) entry_extension = 'a -> GramState.t -> string list * GramState.t
+type ('a, 'b) entry_extension = {
+  eext_fun : 'a -> GramState.t -> string list * GramState.t;
+  eext_eq : 'a -> 'a -> bool;
+}
 
 module EntryInterp = struct type _ t = Ex : ('a, 'b) entry_extension -> ('a * 'b) t end
 module EntryInterpMap = EntryCommand.Map(EntryInterp)
@@ -397,7 +377,7 @@ let extend_grammar_command tag g =
   | [] -> GramState.empty
   | (_, st) :: _ -> st
   in
-  let (rules, st) = modify g grammar_state in
+  let (rules, st) = modify.gext_fun g grammar_state in
   let () = List.iter iter_extend_sync rules in
   let nb = List.length rules in
   grammar_stack := (GramExt (nb, GrammarCommand.Dyn (tag, g)), st) :: !grammar_stack
@@ -408,7 +388,7 @@ let extend_entry_command (type a) (type b) (tag : (a, b) entry_command) (g : a)
   | [] -> GramState.empty
   | (_, st) :: _ -> st
   in
-  let (names, st) = modify g grammar_state in
+  let (names, st) = modify.eext_fun g grammar_state in
   let entries = List.map (fun name -> Entry.make name) names in
   let iter name e =
     camlp5_state := ByEntry (tag, name, e) :: !camlp5_state;
@@ -435,9 +415,7 @@ let extend_dyn_grammar (e, _) = match e with
 
 (** Registering extra grammar *)
 
-type any_entry = AnyEntry : 'a Entry.t -> any_entry
-
-let grammar_names : any_entry list String.Map.t ref = ref String.Map.empty
+let grammar_names : Entry.any_t list String.Map.t ref = ref String.Map.empty
 
 let register_grammars_by_name name grams =
   grammar_names := String.Map.add name grams !grammar_names
@@ -446,7 +424,7 @@ let find_grammars_by_name name =
   try String.Map.find name !grammar_names
   with Not_found ->
     let fold (EntryDataMap.Any (tag, EntryData.Ex map)) accu =
-      try AnyEntry (String.Map.find name map) :: accu
+      try Entry.Any (String.Map.find name map) :: accu
       with Not_found -> accu
     in
     EntryDataMap.fold fold !camlp5_entries []
@@ -461,10 +439,27 @@ type frozen_t =
 let freeze ~marshallable : frozen_t =
   (!grammar_stack, CLexer.get_keyword_state ())
 
+let eq_grams (g1, _) (g2, _) = match g1, g2 with
+| GramExt (_, GrammarCommand.Dyn (t1, v1)), GramExt (_, GrammarCommand.Dyn (t2, v2)) ->
+  begin match GrammarCommand.eq t1 t2 with
+  | None -> false
+  | Some Refl ->
+    let data = GrammarInterpMap.find t1 !grammar_interp in
+    data.gext_eq v1 v2
+  end
+| EntryExt (_, t1, v1), EntryExt (_, t2, v2) ->
+  begin match EntryCommand.eq t1 t2 with
+  | None -> false
+  | Some Refl ->
+    let EntryInterp.Ex data = EntryInterpMap.find t1 !entry_interp in
+    data.eext_eq v1 v2
+  end
+| (GramExt _, EntryExt _) | (EntryExt _, GramExt _) -> false
+
 (* We compare the current state of the grammar and the state to unfreeze,
    by computing the longest common suffixes *)
 let factorize_grams l1 l2 =
-  if l1 == l2 then ([], [], l1) else List.share_tails l1 l2
+  if l1 == l2 then ([], [], l1) else List.share_tails eq_grams l1 l2
 
 let rec number_of_entries accu = function
 | [] -> accu
@@ -486,7 +481,8 @@ let unfreeze (grams, lex) =
 
 let parser_summary_tag =
   Summary.declare_summary_tag "GRAMMAR_LEXER"
-    { Summary.freeze_function = freeze;
+    { stage = Summary.Stage.Synterp;
+      Summary.freeze_function = freeze;
       Summary.unfreeze_function = unfreeze;
       Summary.init_function = Summary.nop }
 
@@ -506,6 +502,7 @@ let () =
   Grammar.register0 wit_int (Prim.integer);
   Grammar.register0 wit_string (Prim.string);
   Grammar.register0 wit_pre_ident (Prim.preident);
+  Grammar.register0 wit_identref (Prim.identref);
   Grammar.register0 wit_ident (Prim.ident);
   Grammar.register0 wit_hyp (Prim.hyp);
   Grammar.register0 wit_ref (Prim.reference);
