@@ -13,12 +13,12 @@ open CErrors
 open Util
 open Constr
 open Context
+open Termops
 open EConstr
 open Vars
 open Namegen
 open Names
 open Pp
-open Termops
 open Tactics
 open Indfun_common
 open Libnames
@@ -45,10 +45,9 @@ type 'a dynamic_info =
 type body_info = constr dynamic_info
 
 let observe_tac s =
-  New.observe_tac ~header:(str "observation") (fun _ _ -> Pp.str s)
+  observe_tac ~header:(str "observation") (fun _ _ -> Pp.str s)
 
 let finish_proof dynamic_infos = observe_tac "finish" assumption
-let refine c = Logic.refiner ~check:true EConstr.Unsafe.(to_constr c)
 let thin = clear
 let eq_constr sigma u v = EConstr.eq_constr_nounivs sigma u v
 
@@ -122,7 +121,7 @@ let prove_trivial_eq h_id context (constructor, type_of_term, term) =
             :: List.map mkVar context_hyps
           in
           let to_refine = applist (mkVar h_id, List.rev context_hyps') in
-          refine to_refine) ]
+          Tactics.exact_check to_refine) ]
 
 let find_rectype env sigma c =
   let t, l = decompose_app sigma (Reductionops.whd_betaiotazeta env sigma c) in
@@ -131,7 +130,7 @@ let find_rectype env sigma c =
   | Construct _ -> (t, l)
   | _ -> raise Not_found
 
-let isAppConstruct ?(env = Global.env ()) sigma t =
+let isAppConstruct env sigma t =
   try
     let t', l = find_rectype env sigma t in
     observe
@@ -191,7 +190,7 @@ let change_eq env sigma hyp_id (context : rel_context) x t end_of_type =
       with Not_found ->
         assert (closed0 sigma t1);
         Int.Map.add t2 t1 sub )
-    else if isAppConstruct sigma t1 && isAppConstruct sigma t2 then begin
+    else if isAppConstruct env sigma t1 && isAppConstruct env sigma t2 then begin
       let c1, args1 = find_rectype env sigma t1
       and c2, args2 = find_rectype env sigma t2 in
       if not (eq_constr c1 c2) then nochange "cannot solve (diff)";
@@ -264,7 +263,7 @@ let change_eq env sigma hyp_id (context : rel_context) x t end_of_type =
              Typing.type_of (Proofview.Goal.env g) (Proofview.Goal.sigma g)
                to_refine
            in
-           tclTHEN (Proofview.Unsafe.tclEVARS evm) (refine to_refine)))
+           tclTHEN (Proofview.Unsafe.tclEVARS evm) (Tactics.exact_check to_refine)))
   in
   let simpl_eq_tac =
     change_hyp_with_using "prove_pattern_simplification" hyp_id new_type_of_hyp
@@ -311,8 +310,9 @@ let rewrite_until_var arg_num eq_ids : unit Proofview.tactic =
   *)
   let test_var g =
     let sigma = Proofview.Goal.sigma g in
+    let env = Proofview.Goal.env g in
     let _, args = destApp sigma (Proofview.Goal.concl g) in
-    not (isConstruct sigma args.(arg_num) || isAppConstruct sigma args.(arg_num))
+    not (isConstruct sigma args.(arg_num) || isAppConstruct env sigma args.(arg_num))
   in
   let rec do_rewrite eq_ids =
     Proofview.Goal.enter (fun g ->
@@ -390,7 +390,7 @@ let clean_hyp_with_heq ptes_infos eq_hyps hyp_id env sigma =
                     [ (* observe_tac "prove rec hyp" *)
                       prove_rec_hyp eq_hyps
                     ; (*                      observe_tac "prove rec hyp" *)
-                      refine to_refine ]) ]
+                      Tactics.exact_check to_refine ]) ]
         in
         tclTHENLIST
           [ (*              observe_tac "hyp rec"  *)
@@ -428,7 +428,7 @@ let clean_hyp_with_heq ptes_infos eq_hyps hyp_id env sigma =
                       ( mkVar hyp_id
                       , List.rev (coq_I :: List.map mkVar context_hyps) )
                   in
-                  refine to_refine) ]
+                  Tactics.exact_check to_refine) ]
         in
         tclTHENLIST
           [ change_hyp_with_using "prove_trivial" hyp_id real_type_of_hyp
@@ -719,7 +719,7 @@ let build_proof (interactive_proof : bool) (fnames : Constant.t list) ptes_infos
         | Array _ -> CErrors.user_err Pp.(str "Arrays not handled yet"))
   and build_proof do_finalize dyn_infos =
     (*     observe (str "proving with "++Printer.pr_lconstr dyn_infos.info++ str " on goal " ++ pr_gls g); *)
-    Indfun_common.New.observe_tac ~header:(str "observation")
+    Indfun_common.observe_tac ~header:(str "observation")
       (fun env sigma ->
         str "build_proof with " ++ pr_leconstr_env env sigma dyn_infos.info)
       (build_proof_aux do_finalize dyn_infos)
@@ -773,7 +773,7 @@ let prove_rec_hyp_for_struct fix_info eq_hyps =
          let rec_hyp_proof =
            mkApp (mkVar fix_info.name, array_get_start pte_args)
          in
-         refine rec_hyp_proof))
+         Tactics.exact_check rec_hyp_proof))
 
 let prove_rec_hyp fix_info =
   {proving_tac = prove_rec_hyp_for_struct fix_info; is_valid = (fun _ -> true)}
@@ -782,7 +782,7 @@ let generalize_non_dep hyp =
   Proofview.Goal.enter (fun g ->
       (*   observe (str "rec id := " ++ Ppconstr.pr_id hyp); *)
       let hyps = [hyp] in
-      let env = Global.env () in
+      let env = Proofview.Goal.env g in
       let sigma = Proofview.Goal.sigma g in
       let hyp_typ = Tacmach.pf_get_hyp_typ hyp g in
       let to_revert, _ =
@@ -799,7 +799,7 @@ let generalize_non_dep hyp =
               (* should be dangerous *)
             then (clear, decl :: keep)
             else (hyp :: clear, keep))
-          ~init:([], []) (Proofview.Goal.env g)
+          ~init:([], []) env
       in
       (*   observe (str "to_revert := " ++ prlist_with_sep spc Ppconstr.pr_id to_revert); *)
       Tacticals.tclTHEN
@@ -813,13 +813,13 @@ let var_of_decl = id_of_decl %> mkVar
 let revert idl =
   Tacticals.tclTHEN (generalize (List.map mkVar idl)) (clear idl)
 
-let generate_equation_lemma evd fnames f fun_num nb_params nb_args rec_args_num
+let generate_equation_lemma env evd fnames f fun_num nb_params nb_args rec_args_num
     =
   let open Tacticals in
   (*   observe (str "nb_args := " ++ str (string_of_int nb_args)); *)
   (*   observe (str "nb_params := " ++ str (string_of_int nb_params)); *)
   (*   observe (str "rec_args_num := " ++ str (string_of_int (rec_args_num + 1) )); *)
-  let f_def = Global.lookup_constant (fst (destConst evd f)) in
+  let f_def = Environ.lookup_constant (fst (destConst evd f)) env in
   let eq_lhs =
     mkApp
       ( f
@@ -846,7 +846,7 @@ let generate_equation_lemma evd fnames f fun_num nb_params nb_args rec_args_num
   in
   (*   observe (str "f_body_with_params_and_other_fun " ++  pr_lconstr f_body_with_params_and_other_fun); *)
   let eq_rhs =
-    Reductionops.nf_betaiotazeta (Global.env ()) evd
+    Reductionops.nf_betaiotazeta env evd
       (mkApp
          ( compose_lam params f_body_with_params_and_other_fun
          , Array.init (nb_params + nb_args) (fun i ->
@@ -854,7 +854,7 @@ let generate_equation_lemma evd fnames f fun_num nb_params nb_args rec_args_num
   in
   (*   observe (str "eq_rhs " ++  pr_lconstr eq_rhs); *)
   let (type_ctxt, type_of_f), evd =
-    let evd, t = Typing.type_of ~refresh:true (Global.env ()) evd f in
+    let evd, t = Typing.type_of ~refresh:true env evd f in
     (decompose_prod_n_assum evd (nb_params + nb_args) t, evd)
   in
   let eqn = mkApp (Lazy.force eq, [|type_of_f; eq_lhs; eq_rhs|]) in
@@ -891,6 +891,7 @@ let generate_equation_lemma evd fnames f fun_num nb_params nb_args rec_args_num
 let do_replace (evd : Evd.evar_map ref) params rec_arg_num rev_args_id f fun_num
     all_funs =
   Proofview.Goal.enter (fun g ->
+      let env = Proofview.Goal.env g in
       let equation_lemma =
         try
           let finfos =
@@ -906,7 +907,7 @@ let do_replace (evd : Evd.evar_map ref) params rec_arg_num rev_args_id f fun_num
             i*)
           let equation_lemma_id = mk_equation_id f_id in
           evd :=
-            generate_equation_lemma !evd all_funs f fun_num (List.length params)
+            generate_equation_lemma env !evd all_funs f fun_num (List.length params)
               (List.length rev_args_id) rec_arg_num;
           let _ =
             match e with
@@ -1142,7 +1143,7 @@ let prove_princ_for_struct (evd : Evd.evar_map ref) interactive_proof fun_num
             if this_fix_info.idx + 1 = 0 then Proofview.tclUNIT ()
               (* Someone  tries to defined a principle on a fully parametric definition declared as a fixpoint (strange but ....) *)
             else
-              Indfun_common.New.observe_tac ~header:(str "observation")
+              Indfun_common.observe_tac ~header:(str "observation")
                 (fun _ _ -> str "h_fix " ++ int (this_fix_info.idx + 1))
                 (fix this_fix_info.name (this_fix_info.idx + 1))
           else
@@ -1308,7 +1309,7 @@ let prove_with_tcc tcc_lemma_constr eqs : unit Proofview.tactic =
         (*               let ids = List.filter (fun id -> not (List.mem id ids)) ids' in  *)
         (*               rewrite *)
         (*            ) *)
-        Eauto.gen_eauto (false, 5) [] (Some []) ]
+        Eauto.gen_eauto ~depth:5 [] (Some []) ]
   | Not_needed -> Proofview.tclUNIT ()
 
 let backtrack_eqs_until_hrec hrec eqs : unit Proofview.tactic =
@@ -1376,7 +1377,7 @@ let new_prove_with_tcc is_mes acc_inv hrec tcc_hyps eqs : unit Proofview.tactic
                    ; observe_tac "rewrite_eqs_in_eqs" (rewrite_eqs_in_eqs eqs)
                    ; observe_tac "finishing using"
                        (tclCOMPLETE
-                          (Eauto.eauto_with_bases (true, 5)
+                          (Eauto.eauto_with_bases ~depth:5
                              [(fun _ sigma -> (sigma, Lazy.force refl_equal))]
                              [Hints.Hint_db.empty TransparentState.empty false]))
                    ]) ] ] ]
