@@ -17,9 +17,12 @@ open Tac2expr
 open Tac2entries.Pltac
 open Proofview.Notations
 
+let ltac2_plugin = "coq-core.plugins.ltac2"
+
 let constr_flags =
   let open Pretyping in
   {
+    use_coercions = true;
     use_typeclasses = Pretyping.UseTC;
     solve_unification_constraints = true;
     fail_evar = true;
@@ -31,6 +34,7 @@ let constr_flags =
 let open_constr_no_classes_flags =
   let open Pretyping in
   {
+  use_coercions = true;
   use_typeclasses = Pretyping.NoUseTC;
   solve_unification_constraints = true;
   fail_evar = false;
@@ -62,6 +66,7 @@ let t_array = coq_core "array"
 let t_unit = coq_core "unit"
 let t_list = coq_core "list"
 let t_constr = coq_core "constr"
+let t_preterm = coq_core "preterm"
 let t_pattern = coq_core "pattern"
 let t_ident = coq_core "ident"
 let t_option = coq_core "option"
@@ -85,6 +90,17 @@ open Core
 let v_unit = Value.of_unit ()
 let v_blk = Valexpr.make_block
 
+let of_relevance = function
+  | Sorts.Relevant -> ValInt 0
+  | Sorts.Irrelevant -> ValInt 1
+
+let to_relevance = function
+  | ValInt 0 -> Sorts.Relevant
+  | ValInt 1 -> Sorts.Irrelevant
+  | _ -> assert false
+
+let relevance = make_repr of_relevance to_relevance
+
 let of_binder b =
   Value.of_ext Value.val_binder b
 
@@ -140,12 +156,19 @@ let err_notfound =
 let err_matchfailure =
   Tac2interp.LtacError (coq_core "Match_failure", [||])
 
+let err_division_by_zero =
+  Tac2interp.LtacError (coq_core "Division_by_zero", [||])
+
 (** Helper functions *)
 
 let thaw f = Tac2ffi.apply f [v_unit]
 
 let fatal_flag : unit Exninfo.t = Exninfo.make ()
 
+let has_fatal_flag info = match Exninfo.get info fatal_flag with
+  | None -> false
+  | Some () -> true
+
 let set_bt info =
   if !Tac2interp.print_ltac2_backtrace then
     Tac2interp.get_backtrace >>= fun bt ->
@@ -162,7 +185,7 @@ let fail ?(info = Exninfo.null) e =
   Proofview.tclZERO ~info e
 
 let return x = Proofview.tclUNIT x
-let pname s = { mltac_plugin = "ltac2"; mltac_tactic = s }
+let pname s = { mltac_plugin = "coq-core.plugins.ltac2"; mltac_tactic = s }
 
 let wrap f =
   return () >>= fun () -> return (f ())
@@ -170,13 +193,33 @@ let wrap f =
 let wrap_unit f =
   return () >>= fun () -> f (); return v_unit
 
+let catchable_exception = function
+  | Logic_monad.Exception _ -> false
+  | e -> CErrors.noncritical e
+
+(* Adds ltac2 backtrace
+   With [passthrough:false], acts like [Proofview.wrap_exceptions] + Ltac2 backtrace handling
+*)
+let wrap_exceptions ?(passthrough=false) f =
+  try f ()
+  with e ->
+    let e, info = Exninfo.capture e in
+    set_bt info >>= fun info ->
+    if not passthrough && catchable_exception e
+    then begin if has_fatal_flag info
+      then Proofview.tclLIFT (Proofview.NonLogical.raise (e, info))
+      else Proofview.tclZERO ~info e
+    end
+    else Exninfo.iraise (e, info)
+
 let assert_focussed =
   Proofview.Goal.goals >>= fun gls ->
   match gls with
   | [_] -> Proofview.tclUNIT ()
   | [] | _ :: _ :: _ -> throw err_notfocussed
 
-let pf_apply f =
+let pf_apply ?(catch_exceptions=false) f =
+  let f env sigma = wrap_exceptions ~passthrough:(not catch_exceptions) (fun () -> f env sigma) in
   Proofview.Goal.goals >>= function
   | [] ->
     Proofview.tclENV >>= fun env ->
@@ -203,6 +246,10 @@ let define2 name r0 r1 f = define_primitive name (arity_suc arity_one) begin fun
   f (Value.repr_to r0 x) (Value.repr_to r1 y)
 end
 
+let define_equality name r eq = define2 name r r begin fun x y ->
+    return (Value.of_bool (eq x y))
+end
+
 let define3 name r0 r1 r2 f = define_primitive name (arity_suc (arity_suc arity_one)) begin fun x y z ->
   f (Value.repr_to r0 x) (Value.repr_to r1 y) (Value.repr_to r2 z)
 end
@@ -226,7 +273,11 @@ let () = define1 "message_of_int" int begin fun n ->
 end
 
 let () = define1 "message_of_string" string begin fun s ->
-  return (Value.of_pp (str (Bytes.to_string s)))
+  return (Value.of_pp (str s))
+end
+
+let () = define1 "message_to_string" pp begin fun pp ->
+  return (Value.of_string (Pp.string_of_ppcmds pp))
 end
 
 let () = define1 "message_of_constr" constr begin fun c ->
@@ -274,7 +325,7 @@ let () = define1 "format_ident" format begin fun s ->
 end
 
 let () = define2 "format_literal" string format begin fun lit s ->
-  return (Value.of_ext val_format (Tac2print.FmtLiteral (Bytes.to_string lit) :: s))
+  return (Value.of_ext val_format (Tac2print.FmtLiteral lit :: s))
 end
 
 let () = define1 "format_alpha" format begin fun s ->
@@ -299,7 +350,7 @@ let () = define2 "format_kfprintf" closure format begin fun k fmt ->
       eval (Pp.app accu (Pp.str s)) args fmt
     | FmtString ->
       let (s, args) = pop1 args in
-      let pp = Pp.str (Bytes.to_string (to_string s)) in
+      let pp = Pp.str (to_string s) in
       eval (Pp.app accu pp) args fmt
     | FmtInt ->
       let (i, args) = pop1 args in
@@ -372,11 +423,11 @@ let () = define2 "ident_equal" ident ident begin fun id1 id2 ->
 end
 
 let () = define1 "ident_to_string" ident begin fun id ->
-  return (Value.of_string (Bytes.of_string (Id.to_string id)))
+  return (Value.of_string (Id.to_string id))
 end
 
 let () = define1 "ident_of_string" string begin fun s ->
-  let id = try Some (Id.of_string (Bytes.to_string s)) with _ -> None in
+  let id = try Some (Id.of_string s) with _ -> None in
   return (Value.of_option Value.of_ident id)
 end
 
@@ -386,6 +437,10 @@ let () = define2 "int_equal" int int begin fun m n ->
   return (Value.of_bool (m == n))
 end
 
+let unop n f = define1 n int begin fun m ->
+  return (Value.of_int (f m))
+end
+
 let binop n f = define2 n int int begin fun m n ->
   return (Value.of_int (f m n))
 end
@@ -394,10 +449,24 @@ let () = binop "int_compare" Int.compare
 let () = binop "int_add" (+)
 let () = binop "int_sub" (-)
 let () = binop "int_mul" ( * )
-
-let () = define1 "int_neg" int begin fun m ->
-  return (Value.of_int (~- m))
+let () = define2 "int_div" int int begin fun m n ->
+  if n == 0 then throw err_division_by_zero
+  else return (Value.of_int (m / n))
 end
+let () = define2 "int_mod" int int begin fun m n ->
+  if n == 0 then throw err_division_by_zero
+  else return (Value.of_int (m mod n))
+end
+let () = unop "int_neg" (~-)
+let () = unop "int_abs" abs
+
+let () = binop "int_asr" (asr)
+let () = binop "int_lsl" (lsl)
+let () = binop "int_lsr" (lsr)
+let () = binop "int_land" (land)
+let () = binop "int_lor" (lor)
+let () = binop "int_lxor" (lxor)
+let () = unop "int_lnot" lnot
 
 (** Char *)
 
@@ -413,34 +482,49 @@ end
 
 let () = define2 "string_make" int char begin fun n c ->
   if n < 0 || n > Sys.max_string_length then throw err_outofbounds
-  else wrap (fun () -> Value.of_string (Bytes.make n c))
+  else wrap (fun () -> Value.of_bytes (Bytes.make n c))
 end
 
-let () = define1 "string_length" string begin fun s ->
+let () = define1 "string_length" bytes begin fun s ->
   return (Value.of_int (Bytes.length s))
 end
 
-let () = define3 "string_set" string int char begin fun s n c ->
+let () = define3 "string_set" bytes int char begin fun s n c ->
   if n < 0 || n >= Bytes.length s then throw err_outofbounds
   else wrap_unit (fun () -> Bytes.set s n c)
 end
 
-let () = define2 "string_get" string int begin fun s n ->
+let () = define2 "string_get" bytes int begin fun s n ->
   if n < 0 || n >= Bytes.length s then throw err_outofbounds
   else wrap (fun () -> Value.of_char (Bytes.get s n))
 end
 
+let () = define2 "string_concat" bytes (list bytes) begin fun sep l ->
+    return (Value.of_bytes (Bytes.concat sep l))
+  end
+
+let () = define2 "string_app" bytes bytes begin fun a b ->
+    return (Value.of_bytes (Bytes.concat Bytes.empty [a; b]))
+  end
+
+let () = define2 "string_equal" bytes bytes begin fun a b ->
+    return (Value.of_bool (Bytes.equal a b))
+  end
+
+let () = define2 "string_compare" bytes bytes begin fun a b ->
+    return (Value.of_int (Bytes.compare a b))
+  end
+
 (** Terms *)
 
 (** constr -> constr *)
 let () = define1 "constr_type" constr begin fun c ->
   let get_type env sigma =
-  Proofview.V82.wrap_exceptions begin fun () ->
     let (sigma, t) = Typing.type_of env sigma c in
     let t = Value.of_constr t in
     Proofview.Unsafe.tclEVARS sigma <*> Proofview.tclUNIT t
-  end in
-  pf_apply get_type
+  in
+  pf_apply ~catch_exceptions:true get_type
 end
 
 (** constr -> constr *)
@@ -462,8 +546,9 @@ let () = define1 "constr_kind" constr begin fun c ->
   | Meta n ->
     v_blk 2 [|Value.of_int n|]
   | Evar (evk, args) ->
+    let args = Evd.expand_existential sigma (evk, args) in
     v_blk 3 [|
-      Value.of_int (Evar.repr evk);
+      Value.of_evar evk;
       Value.of_array Value.of_constr (Array.of_list args);
     |]
   | Sort s ->
@@ -563,9 +648,9 @@ let () = define1 "constr_make" valexpr begin fun knd ->
     let n = Value.to_int n in
     EConstr.mkMeta n
   | (3, [|evk; args|]) ->
-    let evk = Evar.unsafe_of_int (Value.to_int evk) in
+    let evk = to_evar evk in
     let args = Value.to_array Value.to_constr args in
-    EConstr.mkEvar (evk, Array.to_list args)
+    EConstr.mkLEvar sigma (evk, Array.to_list args)
   | (4, [|s|]) ->
     let s = Value.to_ext Value.val_sort s in
     EConstr.mkSort (EConstr.Unsafe.to_sorts s)
@@ -652,16 +737,34 @@ let () = define1 "constr_check" constr begin fun c ->
   end
 end
 
+let () = define3 "constr_liftn" int int constr begin fun n k c ->
+  let ans = EConstr.Vars.liftn n k c in
+  return (Value.of_constr ans)
+end
+
 let () = define3 "constr_substnl" (list constr) int constr begin fun subst k c ->
   let ans = EConstr.Vars.substnl subst k c in
   return (Value.of_constr ans)
 end
 
 let () = define3 "constr_closenl" (list ident) int constr begin fun ids k c ->
-  let ans = EConstr.Vars.substn_vars k ids c in
+  Proofview.tclEVARMAP >>= fun sigma ->
+  let ans = EConstr.Vars.substn_vars sigma k ids c in
   return (Value.of_constr ans)
 end
 
+let () = define2 "constr_closedn" int constr begin fun n c ->
+  Proofview.tclEVARMAP >>= fun sigma ->
+  let ans = EConstr.Vars.closedn sigma n c in
+  return (Value.of_bool ans)
+end
+
+let () = define3 "constr_occur_between" int int constr begin fun n m c ->
+  Proofview.tclEVARMAP >>= fun sigma ->
+  let ans = EConstr.Vars.noccur_between sigma n m c in
+  return (Value.of_bool (not ans))
+end
+
 let () = define1 "constr_case" (repr_ext val_inductive) begin fun ind ->
   Proofview.tclENV >>= fun env ->
   try
@@ -690,7 +793,7 @@ let () = define3 "constr_in_context" ident constr closure begin fun id t c ->
     let sigma = Proofview.Goal.sigma gl in
     let has_var =
       try
-        let _ = Environ.lookup_named_val id env in
+        let _ = Environ.lookup_named id env in
         true
       with Not_found -> false
     in
@@ -698,17 +801,23 @@ let () = define3 "constr_in_context" ident constr closure begin fun id t c ->
       Tacticals.tclZEROMSG (str "Variable already exists")
     else
       let open Context.Named.Declaration in
-      let nenv = EConstr.push_named (LocalAssum (Context.make_annot id Sorts.Relevant, t)) env in
+      let sigma, t_rel =
+        let t_ty = Retyping.get_type_of env sigma t in
+        (* If the user passed eg ['_] for the type we force it to indeed be a type *)
+        let sigma, j = Typing.type_judgment env sigma {uj_val=t; uj_type=t_ty} in
+        sigma, Sorts.relevance_of_sort j.utj_type
+      in
+      let nenv = EConstr.push_named (LocalAssum (Context.make_annot id t_rel, t)) env in
       let (sigma, (evt, _)) = Evarutil.new_type_evar nenv sigma Evd.univ_flexible in
       let (sigma, evk) = Evarutil.new_pure_evar (Environ.named_context_val nenv) sigma evt in
       Proofview.Unsafe.tclEVARS sigma >>= fun () ->
       Proofview.Unsafe.tclSETGOALS [Proofview.with_empty_state evk] >>= fun () ->
       thaw c >>= fun _ ->
       Proofview.Unsafe.tclSETGOALS [Proofview.with_empty_state (Proofview.Goal.goal gl)] >>= fun () ->
-      let args = List.map (fun d -> EConstr.mkVar (get_id d)) (EConstr.named_context env) in
-      let args = EConstr.mkRel 1 :: args in
+      let args = EConstr.identity_subst_val (Environ.named_context_val env) in
+      let args = SList.cons (EConstr.mkRel 1) args in
       let ans = EConstr.mkEvar (evk, args) in
-      let ans = EConstr.mkLambda (Context.make_annot (Name id) Sorts.Relevant, t, ans) in
+      let ans = EConstr.mkLambda (Context.make_annot (Name id) t_rel, t, ans) in
       return (Value.of_constr ans)
   | _ ->
     throw err_notfocussed
@@ -716,34 +825,42 @@ end
 
 (** preterm -> constr *)
 let () = define1 "constr_pretype" (repr_ext val_preterm) begin fun c ->
-  let open Pretyping in
-  let open Ltac_pretype in
-  let pretype env sigma =
-  Proofview.V82.wrap_exceptions begin fun () ->
-    (* For now there are no primitives to create preterms with a non-empty
-       closure. I do not know whether [closed_glob_constr] is really the type
-       we want but it does not hurt in the meantime. *)
-    let { closure; term } = c in
-    let vars = {
-      ltac_constrs = closure.typed;
-      ltac_uconstrs = closure.untyped;
-      ltac_idents = closure.idents;
-      ltac_genargs = Id.Map.empty;
-    } in
-    let flags = constr_flags in
-    let sigma, t = understand_ltac flags env sigma vars WithoutTypeConstraint term in
-    let t = Value.of_constr t in
-    Proofview.Unsafe.tclEVARS sigma <*> Proofview.tclUNIT t
-  end in
-  pf_apply pretype
+    let open Pretyping in
+    let open Ltac_pretype in
+    let pretype env sigma =
+      (* For now there are no primitives to create preterms with a non-empty
+         closure. I do not know whether [closed_glob_constr] is really the type
+         we want but it does not hurt in the meantime. *)
+      let { closure; term } = c in
+      let vars = {
+        ltac_constrs = closure.typed;
+        ltac_uconstrs = closure.untyped;
+        ltac_idents = closure.idents;
+        ltac_genargs = Id.Map.empty;
+      } in
+      let flags = constr_flags in
+      let sigma, t = understand_ltac flags env sigma vars WithoutTypeConstraint term in
+      let t = Value.of_constr t in
+      Proofview.Unsafe.tclEVARS sigma <*> Proofview.tclUNIT t
+    in
+    pf_apply ~catch_exceptions:true pretype
 end
 
 let () = define2 "constr_binder_make" (option ident) constr begin fun na ty ->
   pf_apply begin fun env sigma ->
-  let rel = Retyping.relevance_of_type env sigma ty in
+    match Retyping.relevance_of_type env sigma ty with
+    | rel ->
+      let na = match na with None -> Anonymous | Some id -> Name id in
+      return (Value.of_ext val_binder (Context.make_annot na rel, ty))
+    | exception (Retyping.RetypeError _ as e) ->
+      let e, info = Exninfo.capture e in
+      fail ~info (CErrors.UserError Pp.(str "Not a type."))
+  end
+end
+
+let () = define3 "constr_binder_unsafe_make" (option ident) relevance constr begin fun na rel ty ->
   let na = match na with None -> Anonymous | Some id -> Name id in
   return (Value.of_ext val_binder (Context.make_annot na rel, ty))
-  end
 end
 
 let () = define1 "constr_binder_name" (repr_ext val_binder) begin fun (bnd, _) ->
@@ -761,12 +878,26 @@ let () = define1 "constr_has_evar" constr begin fun c ->
   Proofview.tclUNIT (Value.of_bool b)
 end
 
+(** Extra equalities *)
+
+let () = define_equality "evar_equal" evar Evar.equal
+let () = define_equality "float_equal" float Float64.equal
+let () = define_equality "uint63_equal" uint63 Uint63.equal
+let () = define_equality "meta_equal" int Int.equal
+
+let () = define_equality "constant_equal" constant Constant.UserOrd.equal
+let () = define_equality "constr_case_equal" (repr_ext val_case) begin fun x y ->
+    Ind.UserOrd.equal x.ci_ind y.ci_ind && Sorts.relevance_equal x.ci_relevance y.ci_relevance
+end
+let () = define_equality "constructor_equal" (repr_ext val_constructor) Construct.UserOrd.equal
+let () = define_equality "projection_equal" (repr_ext val_projection) Projection.UserOrd.equal
+
 (** Patterns *)
 
-let empty_context = EConstr.mkMeta Constr_matching.special_meta
+let empty_context = Constr_matching.empty_context
 
 let () = define0 "pattern_empty_context" begin
-  return (Value.of_constr empty_context)
+  return (Value.of_ext val_matching_context empty_context)
 end
 
 let () = define2 "pattern_matches" pattern constr begin fun pat c ->
@@ -792,10 +923,11 @@ let () = define2 "pattern_matches_subterm" pattern constr begin fun pat c ->
   | IStream.Cons ({ m_sub = (_, sub); m_ctx }, s) ->
     let ans = Id.Map.bindings sub in
     let of_pair (id, c) = Value.of_tuple [| Value.of_ident id; Value.of_constr c |] in
-    let ans = Value.of_tuple [| Value.of_constr (Lazy.force m_ctx); Value.of_list of_pair ans |] in
+    let ans = Value.of_tuple [| Value.of_ext val_matching_context m_ctx; Value.of_list of_pair ans |] in
     Proofview.tclOR (return ans) (fun _ -> of_ans s)
   in
   pf_apply begin fun env sigma ->
+    let pat = Constr_matching.instantiate_pattern env sigma Id.Map.empty pat in
     let ans = Constr_matching.match_subterm env sigma (Id.Set.empty,pat) c in
     of_ans ans
   end
@@ -824,39 +956,47 @@ let () = define2 "pattern_matches_subterm_vect" pattern constr begin fun pat c -
   | IStream.Cons ({ m_sub = (_, sub); m_ctx }, s) ->
     let ans = Id.Map.bindings sub in
     let ans = Array.map_of_list snd ans in
-    let ans = Value.of_tuple [| Value.of_constr (Lazy.force m_ctx); Value.of_array Value.of_constr ans |] in
+    let ans = Value.of_tuple [| Value.of_ext val_matching_context m_ctx; Value.of_array Value.of_constr ans |] in
     Proofview.tclOR (return ans) (fun _ -> of_ans s)
   in
   pf_apply begin fun env sigma ->
+    let pat = Constr_matching.instantiate_pattern env sigma Id.Map.empty pat in
     let ans = Constr_matching.match_subterm env sigma (Id.Set.empty,pat) c in
     of_ans ans
   end
 end
 
-let () = define3 "pattern_matches_goal" bool (list (pair bool pattern)) (pair bool pattern) begin fun rev hp cp ->
+let match_pattern = map_repr
+    (fun (b,pat) -> if b then Tac2match.MatchPattern pat else Tac2match.MatchContext pat)
+    (function Tac2match.MatchPattern pat -> (true, pat) | MatchContext pat -> (false, pat))
+    (pair bool pattern)
+
+let () = define3 "pattern_matches_goal" bool
+    (list (pair (option match_pattern) match_pattern))
+    match_pattern
+  begin fun rev hp cp ->
   assert_focussed >>= fun () ->
   Proofview.Goal.enter_one begin fun gl ->
   let env = Proofview.Goal.env gl in
   let sigma = Proofview.Goal.sigma gl in
   let concl = Proofview.Goal.concl gl in
-  let mk_pattern (b, pat) = if b then Tac2match.MatchPattern pat else Tac2match.MatchContext pat in
-  let r = (List.map mk_pattern hp, mk_pattern cp) in
-  Tac2match.match_goal env sigma concl ~rev r >>= fun (hyps, ctx, subst) ->
-    let of_ctxopt ctx = Value.of_constr (Option.default empty_context ctx) in
-    let hids = Value.of_array Value.of_ident (Array.map_of_list fst hyps) in
-    let hctx = Value.of_array of_ctxopt (Array.map_of_list snd hyps) in
+  Tac2match.match_goal env sigma concl ~rev (hp, cp) >>= fun (hyps, ctx, subst) ->
+    let of_ctxopt ctx = Value.of_ext val_matching_context (Option.default empty_context ctx) in
+    let hids = Value.of_array Value.of_ident (Array.map_of_list pi1 hyps) in
+    let hbctx = Value.of_array of_ctxopt
+        (Array.of_list (CList.filter_map (fun (_,bctx,_) -> bctx) hyps))
+    in
+    let hctx = Value.of_array of_ctxopt (Array.map_of_list pi3 hyps) in
     let subs = Value.of_array Value.of_constr (Array.map_of_list snd (Id.Map.bindings subst)) in
     let cctx = of_ctxopt ctx in
-    let ans = Value.of_tuple [| hids; hctx; subs; cctx |] in
+    let ans = Value.of_tuple [| hids; hbctx; hctx; subs; cctx |] in
     Proofview.tclUNIT ans
   end
 end
 
-let () = define2 "pattern_instantiate" constr constr begin fun ctx c ->
-  let ctx = EConstr.Unsafe.to_constr ctx in
-  let c = EConstr.Unsafe.to_constr c in
-  let ans = Termops.subst_meta [Constr_matching.special_meta, c] ctx in
-  return (Value.of_constr (EConstr.of_constr ans))
+let () = define2 "pattern_instantiate" (repr_ext val_matching_context) constr begin fun ctx c ->
+  let ans = Constr_matching.instantiate_context ctx c in
+  return (Value.of_constr ans)
 end
 
 (** Error *)
@@ -931,8 +1071,7 @@ let () = define0 "shelve_unifiable" begin
   Proofview.shelve_unifiable >>= fun () -> return v_unit
 end
 
-let () = define1 "new_goal" int begin fun ev ->
-  let ev = Evar.unsafe_of_int ev in
+let () = define1 "new_goal" evar begin fun ev ->
   Proofview.tclEVARMAP >>= fun sigma ->
   if Evd.mem sigma ev then
     Proofview.Unsafe.tclNEWGOALS [Proofview.with_empty_state ev] <*> Proofview.tclUNIT v_unit
@@ -1002,7 +1141,6 @@ let () = define2 "abstract" (option ident) closure begin fun id f ->
 end
 
 let () = define2 "time" (option string) closure begin fun s f ->
-  let s = Option.map Bytes.to_string s in
   Proofview.tclTIME s (thaw f)
 end
 
@@ -1220,25 +1358,14 @@ let intern_constr self ist c =
   let (_, (c, _)) = Genintern.intern Stdarg.wit_constr ist c in
   (GlbVal c, gtypref t_constr)
 
-let catchable_exception = function
-  | Logic_monad.Exception _ -> false
-  | e -> CErrors.noncritical e
-
 let interp_constr flags ist c =
   let open Pretyping in
   let ist = to_lvar ist in
-  pf_apply begin fun env sigma ->
-  try
+  pf_apply ~catch_exceptions:true begin fun env sigma ->
     let (sigma, c) = understand_ltac flags env sigma ist WithoutTypeConstraint c in
     let c = Value.of_constr c in
     Proofview.Unsafe.tclEVARS sigma >>= fun () ->
     Proofview.tclUNIT c
-  with e when catchable_exception e ->
-    let (e, info) = Exninfo.capture e in
-    set_bt info >>= fun info ->
-    match Exninfo.get info fatal_flag with
-    | None -> Proofview.tclZERO ~info e
-    | Some () -> throw ~info e
   end
 
 let () =
@@ -1302,6 +1429,10 @@ let () =
   define_ml_object Tac2quote.wit_pattern obj
 
 let () =
+  let intern self ist c =
+    let (_, (c, _)) = Genintern.intern Stdarg.wit_constr ist c in
+    (GlbVal c, gtypref t_preterm)
+  in
   let interp _ c =
     let open Ltac_pretype in
     let closure = {
@@ -1316,7 +1447,7 @@ let () =
   let subst subst c = Detyping.subst_glob_constr (Global.env()) subst c in
   let print env sigma c = str "preterm:(" ++ Printer.pr_lglob_constr_env env sigma c ++ str ")" in
   let obj = {
-    ml_intern = (fun _ _ e -> Empty.abort e);
+    ml_intern = intern;
     ml_interp = interp;
     ml_subst = subst;
     ml_print = print;
@@ -1472,16 +1603,21 @@ let () =
 let () =
   let interp ?loc ~poly env sigma tycon id =
     let ist = Tac2interp.get_env @@ GlobEnv.lfun env in
+    let env = GlobEnv.renamed_env env in
     let c = Id.Map.find id ist.env_ist in
     let c = Value.to_constr c in
-    let t = Retyping.get_type_of (GlobEnv.renamed_env env) sigma c in
+    let t = Retyping.get_type_of env sigma c in
+    let j = { Environ.uj_val = c; uj_type = t } in
     match tycon with
     | None ->
-      { Environ.uj_val = c; Environ.uj_type = t }, sigma
-    | Some ty ->
-      let sigma = Evarconv.unify_leq_delay (GlobEnv.renamed_env env) sigma t ty in
-      let j = { Environ.uj_val = c; Environ.uj_type = ty } in
       j, sigma
+    | Some ty ->
+      let sigma =
+        try Evarconv.unify_leq_delay env sigma t ty
+        with Evarconv.UnableToUnify (sigma,e) ->
+          Pretype_errors.error_actual_type ?loc env sigma j ty e
+      in
+      {j with Environ.uj_type = ty}, sigma
   in
   GlobEnv.register_constr_interp0 wit_ltac2_quotation interp
 
@@ -1519,7 +1655,7 @@ let () =
 let () =
   let e = Tac2entries.Pltac.tac2expr_in_env in
   let inject (loc, v) = Ltac_plugin.Tacexpr.TacGeneric (Some "ltac2", in_gen (rawwit wit_ltac2) v) in
-  Ltac_plugin.Tacentries.create_ltac_quotation "ltac2" inject (e, None)
+  Ltac_plugin.Tacentries.create_ltac_quotation ~plugin:ltac2_plugin "ltac2" inject (e, None)
 
 (* Ltac1 runtime representation of Ltac2 closures. *)
 let typ_ltac2 : valexpr Geninterp.Val.typ =
@@ -1551,7 +1687,7 @@ let () =
   in
   let () = Geninterp.register_interp0 wit_ltac2_val interp_fun in
   define1 "ltac1_lambda" valexpr begin fun f ->
-    let body = Tacexpr.TacGeneric (Some "ltac2", in_gen (glbwit wit_ltac2_val) ()) in
+    let body = Tacexpr.TacGeneric (Some "coq-core.plugins.ltac2", in_gen (glbwit wit_ltac2_val) ()) in
     let clos = CAst.make (Tacexpr.TacFun ([Name arg_id], CAst.make (Tacexpr.TacArg body))) in
     let f = Geninterp.Val.inject (Geninterp.Val.Base typ_ltac2) f in
     let lfun = Id.Map.singleton tac_id f in
@@ -1562,7 +1698,7 @@ let () =
 let ltac2_eval =
   let open Ltac_plugin in
   let ml_name = {
-    Tacexpr.mltac_plugin = "ltac2";
+    Tacexpr.mltac_plugin = "coq-core.plugins.ltac2";
     mltac_tactic = "ltac2_eval";
   } in
   let eval_fun args ist = match args with
@@ -1767,6 +1903,16 @@ let () = add_scope "open_constr" (fun arg ->
     Tac2entries.ScopeRule (Pcoq.Symbol.nterm Pcoq.Constr.constr, act)
   )
 
+let () = add_scope "preterm" (fun arg ->
+    let delimiters = List.map (function
+        | SexprRec (_, { v = Some s }, []) -> s
+        | _ -> scope_fail "preterm" arg)
+        arg
+    in
+    let act e = Tac2quote.of_preterm ~delimiters e in
+    Tac2entries.ScopeRule (Pcoq.Symbol.nterm Pcoq.Constr.constr, act)
+  )
+
 let add_expr_scope name entry f =
   add_scope name begin function
   | [] -> Tac2entries.ScopeRule (Pcoq.Symbol.nterm entry, f)
