@@ -68,7 +68,7 @@ let get_typeclasses_iterative_deepening =
     used. *)
 let get_typeclasses_filtered_unification =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
+    ~depr:true
     ~key:["Typeclasses";"Filtered";"Unification"]
     ~value:false
 
@@ -300,7 +300,7 @@ let rec e_trivial_fail_db only_classes db_list local_db secvars =
     end
   in
   let tacl =
-    Eauto.registered_e_assumption ::
+    Eauto.e_assumption ::
     (tclTHEN Tactics.intro trivial_fail :: [trivial_resolve])
   in
   tclSOLVE tacl
@@ -324,7 +324,6 @@ and e_my_find_search db_list local_db secvars hdc complete only_classes env sigm
   in
   let tac_of_hint =
     fun (flags, h) ->
-      let b = FullHint.priority h in
       let p = FullHint.pattern h in
       let name = FullHint.name h in
       let tac = function
@@ -373,15 +372,17 @@ and e_my_find_search db_list local_db secvars hdc complete only_classes env sigm
       in
       let tac = FullHint.run h tac in
       let tac = if complete then Tacticals.tclCOMPLETE tac else tac in
-      let pp =
+      let pp() =
         match p with
         | Some pat when get_typeclasses_filtered_unification () ->
            str " with pattern " ++ Printer.pr_constr_pattern_env env sigma pat
         | _ -> mt ()
       in
-        match FullHint.repr h with
-        | Extern _ -> (tac, b, true, name, lazy (FullHint.print env sigma h ++ pp))
-        | _ -> (tac, b, false, name, lazy (FullHint.print env sigma h ++ pp))
+      let extern = match FullHint.repr h with
+        | Extern _ -> true
+        | _ -> false
+      in
+      (tac, FullHint.priority h, extern, name, lazy (FullHint.print env sigma h ++ pp ()))
   in
   let hint_of_db = hintmap_of env sigma hdc secvars concl in
   let hintl = List.map_filter (fun db -> match hint_of_db db with
@@ -401,13 +402,26 @@ and e_my_find_search db_list local_db secvars hdc complete only_classes env sigm
           m, List.map (fun x -> tac_of_hint (flags, x)) tacs)
         hintl
     in
-    Some hintl
+    let modes, hintl = List.split hintl in
+    let all_mode_match = List.for_all (fun m -> m != NoMode) modes in
+    let hintl = match hintl with
+      (* Optim: only sort if multiple hint sources were involved *)
+      | [hintl] -> hintl
+      | _ ->
+        let hintl = List.flatten hintl in
+        let hintl = List.stable_sort
+            (fun (_, pri1, _, _, _) (_, pri2, _, _, _) -> Int.compare pri1 pri2)
+            hintl
+        in
+        hintl
+    in
+    Some (all_mode_match, hintl)
 
 and e_trivial_resolve db_list local_db secvars only_classes env sigma concl =
   let hd = try Some (decompose_app_bound sigma concl) with Bound -> None in
   try
     (match e_my_find_search db_list local_db secvars hd true only_classes env sigma concl with
-    | Some l -> List.map_append snd l
+    | Some (_,l) -> l
     | None -> [])
   with Not_found -> []
 
@@ -415,7 +429,7 @@ let e_possible_resolve db_list local_db secvars only_classes env sigma concl =
   let hd = try Some (decompose_app_bound sigma concl) with Bound -> None in
   try
     e_my_find_search db_list local_db secvars hd false only_classes env sigma concl
-  with Not_found -> Some []
+  with Not_found -> Some (true, [])
 
 let cut_of_hints h =
   List.fold_left (fun cut db -> PathOr (Hint_db.cut db, cut)) PathEmpty h
@@ -576,9 +590,6 @@ module Search = struct
       occur_existential evd concl
     else true
 
-  let all_mode_match l =
-    List.for_all (fun m -> m != NoMode) l
-
   exception NonStuckFailure
   (* exception Backtrack *)
 
@@ -757,14 +768,12 @@ module Search = struct
             info.search_only_classes env sigma concl with
     | None ->
       Proofview.tclZERO StuckGoal
-    | Some poss ->
+    | Some (all_mode_match, poss) ->
     (* If no goal depends on the solution of this one or the
        instances are irrelevant/assumed to be unique, then
        we don't need to backtrack, as long as no evar appears in the goal
        This is an overapproximation. Evars could appear in this goal only
        and not any other *)
-    let modes, poss = List.split poss in
-    let poss = List.flatten poss in
     let ortac = if backtrack then Proofview.tclOR else Proofview.tclORELSE in
     let idx = ref 1 in
     let foundone = ref false in
@@ -907,7 +916,7 @@ module Search = struct
                solution could be found, we consider it a failed goal, and let
                proof search proceed on the rest of the
                constraints, thus giving a more precise error message. *)
-            if all_mode_match modes &&
+            if all_mode_match &&
               info.search_best_effort then
               Proofview.tclZERO ~info:ie NonStuckFailure
             else Proofview.tclZERO ~info:ie NoApplicableHint
@@ -1022,13 +1031,13 @@ module Search = struct
     let error (e, info) =
       match e with
       | ReachedLimit ->
-        Tacticals.tclFAIL ~info 0 (str"Proof search reached its limit")
+        Tacticals.tclFAIL ~info (str"Proof search reached its limit")
       | NoApplicableHint ->
-        Tacticals.tclFAIL ~info 0 (str"Proof search failed" ++
+        Tacticals.tclFAIL ~info (str"Proof search failed" ++
                                     (if Option.is_empty depth then mt()
                                      else str" without reaching its limit"))
       | Proofview.MoreThanOneSuccess ->
-        Tacticals.tclFAIL ~info 0 (str"Proof search failed: " ++
+        Tacticals.tclFAIL ~info (str"Proof search failed: " ++
                                        str"more than one success found")
       | e -> Proofview.tclZERO ~info e
     in
@@ -1113,7 +1122,9 @@ module Search = struct
           | Some ev' -> Evar.Set.add ev acc
           | None -> acc) (Evar.Set.union goals nongoals) (Evd.get_typeclass_evars evm')
     in
-    let evm' = evars_reset_evd ~with_conv_pbs:true ~with_univs:false evm' evm in
+    (* FIXME: the need to merge metas seems to come from this being called
+       internally from Unification. It should be handled there instead. *)
+    let evm' = Evd.meta_merge (Evd.meta_list evm) (Evd.clear_metas evm') in
     let evm' = Evd.set_typeclass_evars evm' nongoals' in
     let () = ppdebug 1 (fun () ->
         str"New typeclass evars are: " ++
@@ -1200,7 +1211,7 @@ let is_inference_forced p evd ev =
   try
     if Evar.Set.mem ev (Evd.get_typeclass_evars evd) && p ev
     then
-      let (loc, k) = evar_source ev evd in
+      let (loc, k) = evar_source (Evd.find evd ev) in
       match k with
         | Evar_kinds.ImplicitArg (_, _, b) -> b
         | Evar_kinds.QuestionMark _ -> false
@@ -1211,24 +1222,6 @@ let is_inference_forced p evd ev =
 let is_mandatory p comp evd =
   Evar.Set.exists (is_inference_forced p evd) comp
 
-(** In case of unsatisfiable constraints, build a nice error message *)
-
-let error_unresolvable env comp evd =
-  let is_part ev = match comp with
-  | None -> true
-  | Some s -> Evar.Set.mem ev s
-  in
-  let fold ev evi (found, accu) =
-    let ev_class = class_of_constr env evd evi.evar_concl in
-    if not (Option.is_empty ev_class) && is_part ev then
-      (* focus on one instance if only one was searched for *)
-      if not found then (true, Some ev)
-      else (found, None)
-    else (found, accu)
-   in
-  let (_, ev) = Evd.fold_undefined fold evd (true, None) in
-  Pretype_errors.unsatisfiable_constraints env evd ev comp
-
 (** Check if an evar is concerned by the current resolution attempt,
     (and in particular is in the current component).
     Invariant : this should only be applied to undefined evars. *)
@@ -1299,14 +1292,14 @@ let resolve_all_evars depth unique env p oevd do_split fail =
         if fail && (not do_split || is_mandatory (p evd') comp evd')
         then (* Unable to satisfy the constraints. *)
           let comp = if do_split then Some comp else None in
-          error_unresolvable env comp evd'
+          error_unresolvable env evd' comp
         else (* Best effort: use the best found solution on this component *)
           docomp evd' comps
   in docomp oevd split
 
 let initial_select_evars filter =
   fun evd ev evi ->
-    filter ev (Lazy.from_val (snd evi.Evd.evar_source)) &&
+    filter ev (Lazy.from_val (snd (Evd.evar_source evi))) &&
     (* Typeclass evars can contain evars whose conclusion is not
        yet determined to be a class or not. *)
     Typeclasses.is_class_evar evd evi
@@ -1334,7 +1327,7 @@ let solve_inst env evd filter unique split fail =
   sigma
 
 let () =
-  Hook.set Typeclasses.solve_all_instances_hook solve_inst
+  Typeclasses.set_solve_all_instances solve_inst
 
 let resolve_one_typeclass env ?(sigma=Evd.from_env env) concl unique =
   let (term, sigma) = Hints.wrap_hint_warning_fun env sigma begin fun sigma ->
@@ -1342,9 +1335,9 @@ let resolve_one_typeclass env ?(sigma=Evd.from_env env) concl unique =
   let st = Hint_db.transparent_state hints in
   let modes = Hint_db.modes hints in
   let depth = get_typeclasses_depth () in
-  let tac = Search.eauto_tac (modes,st)
+  let tac = Tacticals.tclCOMPLETE (Search.eauto_tac (modes,st)
       ~only_classes:true ~best_effort:false
-      ~depth [hints] ~dep:true
+      ~depth [hints] ~dep:true)
   in
   let entry, pv = Proofview.init sigma [env, concl] in
   let pv =
@@ -1361,7 +1354,7 @@ let resolve_one_typeclass env ?(sigma=Evd.from_env env) concl unique =
   (sigma, term)
 
 let () =
-  Hook.set Typeclasses.solve_one_instance_hook
+  Typeclasses.set_solve_one_instance
     (fun x y z w -> resolve_one_typeclass x ~sigma:y z w)
 
 (** Take the head of the arity of a constr.
@@ -1383,14 +1376,14 @@ let head_of_constr h c =
 let not_evar c =
   Proofview.tclEVARMAP >>= fun sigma ->
   match EConstr.kind sigma c with
-  | Evar _ -> Tacticals.tclFAIL 0 (str"Evar")
+  | Evar _ -> Tacticals.tclFAIL (str"Evar")
   | _ -> Proofview.tclUNIT ()
 
 let is_ground c =
   let open Tacticals in
   Proofview.tclEVARMAP >>= fun sigma ->
   if Evarutil.is_ground_term sigma c then tclIDTAC
-  else tclFAIL 0 (str"Not ground")
+  else tclFAIL (str"Not ground")
 
 let autoapply c i =
   let open Proofview.Notations in
@@ -1408,5 +1401,5 @@ let autoapply c i =
   Clenv.res_pf ~with_evars:true ~with_classes:false ~flags ce <*>
       Proofview.tclEVARMAP >>= (fun sigma ->
       let sigma = Typeclasses.make_unresolvables
-          (fun ev -> Typeclasses.all_goals ev (Lazy.from_val (snd (Evd.find sigma ev).evar_source))) sigma in
+          (fun ev -> Typeclasses.all_goals ev (Lazy.from_val (snd (Evd.evar_source (Evd.find sigma ev))))) sigma in
       Proofview.Unsafe.tclEVARS sigma) end
