@@ -106,7 +106,7 @@ let make_anonymous_conclusion_flexible sigma = function
   | Some (true, s) ->
     (match EConstr.ESorts.kind sigma s with
      | Type u ->
-       (match Univ.universe_level u with
+       (match Univ.Universe.level u with
         | Some u ->
           Evd.make_flexible_variable sigma ~algebraic:true u
         | None -> sigma)
@@ -175,18 +175,28 @@ let interp_cstrs env (sigma, ind_rel) impls params ind arity =
   in
   (sigma, pred ind_rel), (cnames, ctyps, cimpls)
 
-let sign_level env evd sign =
+(** FIXME: This is a horrible hack, use a saner heuristic *)
+let max_sort s1 s2 = match s1, s2 with
+| (SProp, SProp) | (Prop, Prop) | (Set, Set) -> s1
+| (SProp, (Prop | Set | Type _ as s)) | ((Prop | Set | Type _) as s, SProp) -> s
+| (Prop, (Set | Type _ as s)) | ((Set | Type _) as s, Prop) -> s
+| (Set, Type u) | (Type u, Set) -> Sorts.sort_of_univ (Univ.Universe.sup Univ.Universe.type0 u)
+| (Type u, Type v) -> Sorts.sort_of_univ (Univ.Universe.sup u v)
+
+let compute_constructor_level env evd sign =
   fst (List.fold_right
     (fun d (lev,env) ->
       match d with
-      | LocalDef _ -> lev, push_rel d env
+      | LocalDef _ -> lev, EConstr.push_rel d env
       | LocalAssum _ ->
-        let s = Retyping.get_sort_of env evd (EConstr.of_constr (RelDecl.get_type d)) in
-        let u = univ_of_sort s in
-          (Univ.sup u lev, push_rel d env))
-    sign (Univ.Universe.sprop,env))
+        let s = Retyping.get_sort_of env evd (RelDecl.get_type d) in
+          (max_sort s lev, EConstr.push_rel d env))
+    sign (Sorts.sprop,env))
 
-let sup_list min = List.fold_left Univ.sup min
+let sign_level env sigma sign =
+  compute_constructor_level env sigma (EConstr.of_rel_context sign)
+
+let sup_list min = List.fold_left max_sort min
 
 let extract_level env evd min tys =
   let sorts = List.map (fun ty ->
@@ -194,7 +204,9 @@ let extract_level env evd min tys =
       sign_level env evd (LocalAssum (make_annot Anonymous Sorts.Relevant, concl) :: ctx)) tys
   in sup_list min sorts
 
-let is_flexible_sort evd u =
+let is_flexible_sort evd s = match s with
+| Set | Prop | SProp -> false
+| Type u ->
   match Univ.Universe.level u with
   | Some l -> Evd.is_flexible_level evd l
   | None -> false
@@ -220,8 +232,11 @@ where
 *)
 
 let is_direct_sort_constraint s v = match s with
-  | Some u -> Univ.univ_level_mem u v
-  | None -> false
+| None -> false
+| Some u ->
+  match v with
+  | Sorts.Prop | Sorts.Set | Sorts.SProp -> false
+  | Sorts.Type v -> Univ.univ_level_mem u v
 
 let solve_constraints_system levels level_bounds =
   let open Univ in
@@ -231,7 +246,7 @@ let solve_constraints_system levels level_bounds =
       | Some u ->
         (match Universe.level u with
         | Some u -> Some u
-        | _ -> level_bounds.(i) <- Universe.sup level_bounds.(i) u; None)
+        | _ -> level_bounds.(i) <- max_sort level_bounds.(i) (Sorts.sort_of_univ u); None)
       | None -> None)
       levels in
   let v = Array.copy level_bounds in
@@ -260,31 +275,33 @@ let solve_constraints_system levels level_bounds =
   for i=0 to nind-1 do
     for j=0 to nind-1 do
       if not (Int.equal i j) && Int.Set.mem j clos.(i) then
-        (v.(i) <- Universe.sup v.(i) level_bounds.(j));
+        (v.(i) <- max_sort v.(i) level_bounds.(j));
     done;
   done;
   v
 
 let inductive_levels env evd arities inds =
   let destarities = List.map (fun x -> x, Reduction.dest_arity env x) arities in
-  let levels = List.map (fun (x,(ctx,a)) ->
-    if Sorts.is_prop a || Sorts.is_sprop a then None
-    else Some (univ_of_sort a)) destarities
+  let map (x, (ctx, s)) = match s with
+  | Prop | SProp -> None
+  | Set -> Some Univ.Universe.type0
+  | Type u -> Some u
   in
+  let levels = List.map map destarities in
   let cstrs_levels, sizes =
     CList.split (List.map2 (fun (_,tys) (arity,(ctx,du)) ->
         let len = List.length tys in
-        let minlev = Sorts.univ_of_sort du in
+        let minlev = du in
         let minlev =
           if len > 1 && not (is_impredicative_sort env du) then
-            Univ.sup minlev Univ.type0_univ
+            max_sort minlev Sorts.set
           else minlev
         in
         let minlev =
           (* Indices contribute. *)
           if indices_matter env then begin
             let ilev = sign_level env evd ctx in
-            Univ.sup ilev minlev
+            max_sort ilev minlev
           end
           else minlev
         in
@@ -301,7 +318,7 @@ let inductive_levels env evd arities inds =
     CList.fold_left3 (fun (evd, arities) cu (arity,(ctx,du)) len ->
       if is_impredicative_sort env du then
         (* Any product is allowed here. *)
-        evd, (false, arity) :: arities
+        evd, arity :: arities
       else (* If in a predicative sort, or asked to infer the type,
               we take the max of:
               - indices (if in indices-matter mode)
@@ -311,33 +328,38 @@ let inductive_levels env evd arities inds =
         (* Constructors contribute. *)
         let evd =
           if Sorts.is_set du then
-            if not (Evd.check_leq evd cu Univ.type0_univ) then
+            if not (Evd.check_leq evd cu Sorts.set) then
               raise (InductiveError LargeNonPropInductiveNotInType)
             else evd
           else evd
         in
         let evd =
-          if len >= 2 && Univ.is_type0m_univ cu then
+          if len >= 2 && Sorts.is_prop cu then
            (* "Polymorphic" type constraint and more than one constructor,
                should not land in Prop. Add constraint only if it would
                land in Prop directly (no informative arguments as well). *)
             Evd.set_leq_sort env evd Sorts.set du
           else evd
         in
-        let duu = Sorts.univ_of_sort du in
-        let template_prop, evd =
-          if not (Univ.is_small_univ duu) && Univ.Universe.equal cu duu then
-            if is_flexible_sort evd duu && not (Evd.check_leq evd Univ.type0_univ duu)
+        let evd, arity =
+          if not (Sorts.is_small du) && Sorts.equal cu du then
+            if is_flexible_sort evd du && not (Evd.check_leq evd Sorts.set du)
             then if Term.isArity arity
             (* If not a syntactic arity, the universe may be used in a
                polymorphic instance and so cannot be lowered to Prop.
                See #13300. *)
-              then true, Evd.set_eq_sort env evd Sorts.prop du
-              else false, Evd.set_eq_sort env evd Sorts.set du
-            else false, evd
-          else false, Evd.set_eq_sort env evd (sort_of_univ cu) du
+              then
+                (* Workaround: the kernel does not handle non-Prop unbounded
+                   arities. In this situation we have no constraints from the
+                   constructor so we cook up a new type and unify the unbound
+                   universe to a dummy value. *)
+                let evd = Evd.set_eq_sort env evd Sorts.set du in
+                evd, Term.mkArity (ctx, Sorts.prop)
+              else Evd.set_eq_sort env evd Sorts.set du, arity
+            else evd, arity
+          else Evd.set_eq_sort env evd cu du, arity
         in
-          (evd, (template_prop, arity) :: arities))
+          (evd, arity :: arities))
     (evd,[]) (Array.to_list levels') destarities sizes
   in evd, List.rev arities
 
@@ -347,22 +369,128 @@ let check_named {CAst.loc;v=na} = match na with
   let msg = str "Parameters must be named." in
   user_err ?loc  msg
 
-let template_polymorphism_candidate ~ctor_levels uctx params concl =
-  match uctx with
-  | UState.Monomorphic_entry uctx ->
-    let concltemplate = Option.cata (fun s -> not (Sorts.is_small s)) false concl in
-    if not concltemplate then false
-    else
-      let conclu = Option.cata Sorts.univ_of_sort Univ.type0m_univ concl in
-      Option.has_some @@ IndTyping.template_polymorphic_univs ~ctor_levels uctx params conclu
-  | UState.Polymorphic_entry _ -> false
+(* Returns the list [x_1, ..., x_n] of levels contributing to template
+   polymorphism. The elements x_k is None if the k-th parameter
+   (starting from the most recent and ignoring let-definitions) is not
+   contributing to the inductive type's sort or is Some u_k if its level
+   is u_k and is contributing. *)
+let template_polymorphic_univs ~ctor_levels uctx paramsctxt u =
+  let unbounded_from_below u cstrs =
+    let open Univ in
+    Univ.Constraints.for_all (fun (l, d, r) ->
+        match d with
+        | Eq -> not (Univ.Level.equal l u) && not (Univ.Level.equal r u)
+        | Lt | Le -> not (Univ.Level.equal r u))
+      cstrs
+  in
+  let fold_params accu decl = match decl with
+  | LocalAssum (_, p) ->
+    let c = Term.strip_prod_assum p in
+    begin match Constr.kind c with
+    | Constr.Sort (Type u) ->
+      begin match Univ.Universe.level u with
+      | Some l -> Univ.Level.Set.add l accu
+      | None -> accu
+      end
+    | _ -> accu
+    end
+  | LocalDef _ -> accu
+  in
+  let paramslevels = List.fold_left fold_params Univ.Level.Set.empty paramsctxt in
+  let check_level l =
+    Univ.Level.Set.mem l (Univ.ContextSet.levels uctx) &&
+    Univ.Level.Set.mem l paramslevels &&
+    (let () = assert (not @@ Univ.Level.is_set l) in true) &&
+    unbounded_from_below l (Univ.ContextSet.constraints uctx) &&
+    not (Univ.Level.Set.mem l ctor_levels)
+  in
+  let univs = Univ.Universe.levels u in
+  let univs = Univ.Level.Set.filter (fun l -> check_level l) univs in
+  univs
+
+let template_polymorphism_candidate uctx params entry concl = match concl with
+| None -> Univ.Level.Set.empty
+| Some (Set | SProp | Prop) -> Univ.Level.Set.empty
+| Some (Type u) ->
+  let ctor_levels =
+    let add_levels c levels = Univ.Level.Set.union levels (Vars.universes_of_constr c) in
+    let param_levels =
+      List.fold_left (fun levels d -> match d with
+          | LocalAssum _ -> levels
+          | LocalDef (_,b,t) -> add_levels b (add_levels t levels))
+        Univ.Level.Set.empty params
+    in
+    List.fold_left (fun levels c -> add_levels c levels)
+      param_levels entry.mind_entry_lc
+  in
+  let univs = template_polymorphic_univs ~ctor_levels uctx params u in
+  univs
+
+let split_universe_context subset (univs, csts) =
+  let subfilter (l, _, r) =
+    let () = assert (not @@ Univ.Level.Set.mem r subset) in
+    Univ.Level.Set.mem l subset
+  in
+  let subcst = Univ.Constraints.filter subfilter csts in
+  let rem = Univ.Level.Set.diff univs subset in
+  let remfilter (l, _, r) =
+    not (Univ.Level.Set.mem l subset) && not (Univ.Level.Set.mem r subset)
+  in
+  let remcst = Univ.Constraints.filter remfilter csts in
+  (subset, subcst), (rem, remcst)
+
+let warn_no_template_universe =
+  CWarnings.create ~name:"no-template-universe" ~category:"universes"
+    (fun () -> Pp.str "This inductive type has no template universes.")
+
+let compute_template_inductive ~user_template ~env_ar_params ~ctx_params ~univ_entry entry concl =
+match user_template, univ_entry with
+| Some false, UState.Monomorphic_entry uctx ->
+  Monomorphic_ind_entry, uctx
+| Some false, UState.Polymorphic_entry uctx ->
+  Polymorphic_ind_entry uctx, Univ.ContextSet.empty
+| Some true, UState.Monomorphic_entry uctx ->
+  let template_universes = template_polymorphism_candidate uctx ctx_params entry concl in
+  let template, global = split_universe_context template_universes uctx in
+  let () = if Univ.Level.Set.is_empty (fst template) then warn_no_template_universe () in
+  Template_ind_entry template, global
+| Some true, UState.Polymorphic_entry _ ->
+  user_err Pp.(strbrk "Template-polymorphism and universe polymorphism are not compatible.")
+| None, UState.Polymorphic_entry uctx ->
+  Polymorphic_ind_entry uctx, Univ.ContextSet.empty
+| None, UState.Monomorphic_entry uctx ->
+  (* Heuristic: the user has not written Prop explicitly in the return
+      arity, but inference has decided to lower it to Prop. *)
+  let templatearity =
+    if Term.isArity entry.mind_entry_arity then
+      let (_, s) = Reduction.dest_arity env_ar_params entry.mind_entry_arity in
+      if Sorts.is_prop s then match concl with
+      | None | Some (Type _ | Set)-> true
+      | Some Prop -> false
+      | Some SProp -> assert false
+      else false
+    else false
+  in
+  if templatearity then
+    let template = should_auto_template entry.mind_entry_typename true in
+    (* Dummy template inductive. Matters for the shape of the induction principle *)
+    if template then Template_ind_entry Univ.ContextSet.empty, uctx
+    else Monomorphic_ind_entry, uctx
+  else
+    let template_candidate = template_polymorphism_candidate uctx ctx_params entry concl in
+    let has_template = not @@ Univ.Level.Set.is_empty template_candidate in
+    let template = should_auto_template entry.mind_entry_typename has_template in
+    if template then
+      let template, global = split_universe_context template_candidate uctx in
+      Template_ind_entry template, global
+    else Monomorphic_ind_entry, uctx
 
 let check_param = function
 | CLocalDef (na, _, _) -> check_named na
 | CLocalAssum (nas, Default _, _) -> List.iter check_named nas
 | CLocalAssum (nas, Generalized _, _) -> ()
 | CLocalPattern {CAst.loc} ->
-    Loc.raise ?loc (Stream.Error "pattern with quote not allowed here")
+    Loc.raise ?loc (Gramlib.Stream.Error "pattern with quote not allowed here")
 
 let restrict_inductive_universes sigma ctx_params arities constructors =
   let merge_universes_of_constr c =
@@ -402,55 +530,33 @@ let interp_mutual_inductive_constr ~sigma ~template ~udecl ~variances ~ctx_param
   let sigma, arities = inductive_levels env_ar_params sigma arities constructors in
   let sigma = Evd.minimize_universes sigma in
   let nf = Evarutil.nf_evars_universes sigma in
-  let arities = List.map (on_snd nf) arities in
+  let arities = List.map nf arities in
   let constructors = List.map (on_snd (List.map nf)) constructors in
   let ctx_params = List.map Termops.(map_rel_decl (EConstr.to_constr sigma)) ctx_params in
   let arityconcl = List.map (Option.map (fun (_anon, s) -> EConstr.ESorts.kind sigma s)) arityconcl in
-  let sigma = restrict_inductive_universes sigma ctx_params (List.map snd arities) constructors in
+  let sigma = restrict_inductive_universes sigma ctx_params arities constructors in
   let univ_entry, binders = Evd.check_univ_decl ~poly sigma udecl in
 
   (* Build the inductive entries *)
-  let entries = List.map4 (fun indname (templatearity, arity) concl (cnames,ctypes) ->
+  let entries = List.map3 (fun indname arity (cnames,ctypes) ->
       { mind_entry_typename = indname;
         mind_entry_arity = arity;
         mind_entry_consnames = cnames;
         mind_entry_lc = ctypes
       })
-      indnames arities arityconcl constructors
+      indnames arities constructors
   in
-  let template = List.map4 (fun indname (templatearity, _) concl (_, ctypes) ->
-      let template_candidate () =
-        templatearity ||
-        let ctor_levels =
-          let add_levels c levels = Univ.Level.Set.union levels (Vars.universes_of_constr c) in
-          let param_levels =
-            List.fold_left (fun levels d -> match d with
-                | LocalAssum _ -> levels
-                | LocalDef (_,b,t) -> add_levels b (add_levels t levels))
-              Univ.Level.Set.empty ctx_params
-          in
-          List.fold_left (fun levels c -> add_levels c levels)
-            param_levels ctypes
-        in
-        template_polymorphism_candidate ~ctor_levels univ_entry ctx_params concl
-      in
-      match template with
-        | Some template ->
-          if poly && template then user_err
-              Pp.(strbrk "Template-polymorphism and universe polymorphism are not compatible.");
-          template
-        | None ->
-          should_auto_template indname (template_candidate ())
-      )
-      indnames arities arityconcl constructors
-  in
-  let is_template = List.for_all (fun t -> t) template in
-  let univ_entry, ctx = match univ_entry with
-  | UState.Monomorphic_entry ctx ->
-    if is_template then Template_ind_entry ctx, Univ.ContextSet.empty
-    else Monomorphic_ind_entry, ctx
-  | UState.Polymorphic_entry uctx ->
-    Polymorphic_ind_entry uctx, Univ.ContextSet.empty
+  let univ_entry, ctx = match entries, arityconcl with
+  | [entry], [concl] ->
+    compute_template_inductive ~user_template:template ~env_ar_params ~ctx_params ~univ_entry entry concl
+  | _ ->
+    let () = match template with
+    | Some true -> user_err Pp.(str "Template-polymorphism not allowed with mutual inductives.")
+    | _ -> ()
+    in
+    match univ_entry with
+    | UState.Monomorphic_entry ctx -> Monomorphic_ind_entry, ctx
+    | UState.Polymorphic_entry uctx -> Polymorphic_ind_entry uctx, Univ.ContextSet.empty
   in
   let variance = variance_of_entry ~cumulative ~variances univ_entry in
   (* Build the mutual inductive entry *)
@@ -514,10 +620,9 @@ let interp_mutual_inductive_gen env0 ~template udecl (uparamsl,paramsl,indl) not
   let indnames = List.map (fun ind -> ind.ind_name) indl in
   let ninds = List.length indl in
 
-  (* In case of template polymorphism, we need to compute more constraints *)
-  let env0 = if poly then env0 else Environ.set_universes_lbound env0 UGraph.Bound.Prop in
-
   let sigma, env_params, (ctx_params, env_uparams, ctx_uparams, userimpls, useruimpls, impls, udecl, variances) =
+    (* In case of template polymorphism, we need to compute more constraints *)
+    let env0 = if poly then env0 else Environ.set_universes_lbound env0 UGraph.Bound.Prop in
     interp_params env0 udecl uparamsl paramsl
   in
 
@@ -608,7 +713,11 @@ let eq_params (up1,p1) (up2,p2) =
 
 let extract_coercions indl =
   let mkqid (_,({CAst.v=id},_)) = qualid_of_ident id in
-  let extract lc = List.filter (fun (iscoe,_) -> iscoe) lc in
+  let iscoe (coe, inst) = match inst with
+    (* remove BackInstanceWarning after deprecation phase *)
+    | Vernacexpr.(NoInstance | BackInstanceWarning) -> coe = Vernacexpr.AddCoercion
+    | _ -> user_err (Pp.str "'::' not allowed in inductives.") in
+  let extract lc = List.filter (fun (coe,_) -> iscoe coe) lc in
   List.map mkqid (List.flatten(List.map (fun (_,_,_,lc) -> extract lc) indl))
 
 let extract_params indl =
@@ -638,33 +747,61 @@ type uniform_inductive_flag =
   | UniformParameters
   | NonUniformParameters
 
-let do_mutual_inductive ~template udecl indl ~cumulative ~poly ?typing_flags ~private_ind ~uniform finite =
-  let (params,indl),coes,ntns = extract_mutual_inductive_declaration_components indl in
-  let ntns = List.map Metasyntax.prepare_where_notation ntns in
+module Mind_decl = struct
+
+type t = {
+  mie : Entries.mutual_inductive_entry;
+  nuparams : int option;
+  univ_binders : UnivNames.universe_binders;
+  implicits : DeclareInd.one_inductive_impls list;
+  uctx : Univ.ContextSet.t;
+  where_notations : Metasyntax.where_decl_notation list;
+  coercions : Libnames.qualid list;
+}
+
+end
+
+let rec count_binder_expr = function
+  | [] -> 0
+  | CLocalAssum(l,_,_) :: rest -> List.length l + count_binder_expr rest
+  | CLocalDef _ :: rest -> 1 + count_binder_expr rest
+  | CLocalPattern {CAst.loc} :: _ ->
+    Loc.raise ?loc (Gramlib.Stream.Error "pattern with quote not allowed here")
+
+let interp_mutual_inductive ~env ~template udecl indl ~cumulative ~poly ?typing_flags ~private_ind ~uniform finite =
+  let (params,indl),coercions,ntns = extract_mutual_inductive_declaration_components indl in
+  let where_notations = List.map Metasyntax.prepare_where_notation ntns in
   (* Interpret the types *)
-  let indl = match params with
-    | uparams, Some params -> (uparams, params, indl)
+  let indl, nuparams = match params with
+    | uparams, Some params -> (uparams, params, indl), Some (count_binder_expr params)
     | params, None -> match uniform with
-      | UniformParameters -> (params, [], indl)
-      | NonUniformParameters -> ([], params, indl)
+      | UniformParameters -> (params, [], indl), Some 0
+      | NonUniformParameters -> ([], params, indl), None
   in
-  let env = Global.env () in
   let env = Environ.update_typing_flags ?typing_flags env in
-  let mie,binders,impls,ctx = interp_mutual_inductive_gen env ~template udecl indl ntns ~cumulative ~poly ~private_ind finite in
+  let mie, univ_binders, implicits, uctx = interp_mutual_inductive_gen env ~template udecl indl where_notations ~cumulative ~poly ~private_ind finite in
+  let open Mind_decl in
+  { mie; nuparams; univ_binders; implicits; uctx; where_notations; coercions }
+
+let do_mutual_inductive ~template udecl indl ~cumulative ~poly ?typing_flags ~private_ind ~uniform finite =
+  let open Mind_decl in
+  let env = Global.env () in
+  let { mie; univ_binders; implicits; uctx; where_notations; coercions } =
+    interp_mutual_inductive ~env ~template udecl indl ~cumulative ~poly ?typing_flags ~private_ind ~uniform finite in
   (* Slightly hackish global universe declaration due to template types. *)
   let binders = match mie.mind_entry_universes with
-  | Monomorphic_ind_entry -> (UState.Monomorphic_entry ctx, binders)
-  | Template_ind_entry ctx -> (UState.Monomorphic_entry ctx, binders)
+  | Monomorphic_ind_entry -> (UState.Monomorphic_entry uctx, univ_binders)
+  | Template_ind_entry ctx -> (UState.Monomorphic_entry ctx, univ_binders)
   | Polymorphic_ind_entry uctx -> (UState.Polymorphic_entry uctx, UnivNames.empty_binders)
   in
   (* Declare the global universes *)
-  DeclareUctx.declare_universe_context ~poly:false ctx;
+  DeclareUctx.declare_universe_context ~poly:false uctx;
   (* Declare the mutual inductive block with its associated schemes *)
-  ignore (DeclareInd.declare_mutual_inductive_with_eliminations ?typing_flags mie binders impls);
+  ignore (DeclareInd.declare_mutual_inductive_with_eliminations ?typing_flags mie binders implicits);
   (* Declare the possible notations of inductive types *)
-  List.iter (Metasyntax.add_notation_interpretation ~local:false (Global.env ())) ntns;
+  List.iter (Metasyntax.add_notation_interpretation ~local:false (Global.env ())) where_notations;
   (* Declare the coercions *)
-  List.iter (fun qid -> ComCoercion.try_add_new_coercion (Nametab.locate qid) ~local:false ~poly) coes
+  List.iter (fun qid -> ComCoercion.try_add_new_coercion (Nametab.locate qid) ~local:false ~poly ~nonuniform:false ~reversible:true) coercions
 
 (** Prepare a "match" template for a given inductive type.
     For each branch of the match, we list the constructor name
@@ -705,3 +842,10 @@ let make_cases ind =
        let consref = GlobRef.ConstructRef (ith_constructor_of_inductive ind (i + 1)) in
        (Libnames.string_of_qualid (Nametab.shortest_qualid_of_global Id.Set.empty consref) :: al') :: l)
     mip.mind_nf_lc []
+
+module Internal =
+struct
+
+let compute_constructor_level = compute_constructor_level
+
+end
