@@ -631,6 +631,8 @@ module Internal = struct
 
   let objVariable = objVariable
 
+  let export_side_effects = export_side_effects
+
 end
 
 let declare_definition_scheme ~internal ~univs ~role ~name c =
@@ -828,7 +830,7 @@ module ProgramDecl = struct
     { prg_cinfo : constr CInfo.t
     ; prg_info : Info.t
     ; prg_opaque : bool
-    ; prg_hook : 'a Hook.g option
+    ; prg_hook : 'a option
     ; prg_body : constr
     ; prg_uctx : UState.t
     ; prg_obligations : obligations
@@ -969,13 +971,6 @@ let shrink_body c ty =
 (* Saving an obligation                                                *)
 (***********************************************************************)
 
-let unfold_entry cst = Hints.HintsUnfoldEntry [Tacred.EvalConstRef cst]
-
-let add_hint local prg cst =
-  (* XXX checking sections here is suspicious but matches historical (unintended?) behaviour *)
-  let locality = if local || Global.sections_are_opened () then Hints.Local else Hints.SuperGlobal in
-  Hints.add_hints ~locality [Id.to_string prg.prg_cinfo.CInfo.name] (unfold_entry cst)
-
 let declare_obligation prg obl ~uctx ~types ~body =
   let poly = prg.prg_info.Info.poly in
   let univs = UState.univ_entry ~poly uctx in
@@ -1000,8 +995,6 @@ let declare_obligation prg obl ~uctx ~types ~body =
         ~kind:Decls.(IsProof Property)
         (DefinitionEntry ce)
     in
-    if not opaque then
-      add_hint (Locality.make_section_locality None) prg constant;
     definition_message obl.obl_name;
     let body =
       match fst univs with
@@ -1031,7 +1024,13 @@ module ProgMap = Id.Map
 
 module State = struct
 
-  type t = t ProgramDecl.t CEphemeron.key ProgMap.t
+  type t = prg_hook ProgramDecl.t CEphemeron.key ProgMap.t
+  and prg_hook = PrgHook of t Hook.g
+
+  let call_prg_hook { prg_hook=hook } x pm =
+    let hook = Option.map (fun (PrgHook h) -> h) hook in
+    Hook.call_g ?hook x pm
+
 
   let empty = ProgMap.empty
 
@@ -1143,7 +1142,7 @@ let declare_definition ~pm prg =
   let kn, uctx = declare_definition_core ~cinfo ~info ~obls ~body ~opaque sigma in
   (* XXX: We call the obligation hook here, by consistency with the
      previous imperative behaviour, however I'm not sure this is right *)
-  let pm = Hook.call_g ?hook:prg.prg_hook
+  let pm = State.call_prg_hook prg
       { Hook.S.uctx; obls; scope = prg.prg_info.Info.scope; dref = kn} pm in
   let pm = progmap_remove pm prg in
   pm, kn
@@ -1229,7 +1228,7 @@ let declare_mutual_definition ~pm l =
   Hook.call ?hook:first.prg_info.Info.hook s_hook;
   (* XXX: We call the obligation hook here, by consistency with the
      previous imperative behaviour, however I'm not sure this is right *)
-  let pm = Hook.call_g ?hook:first.prg_hook s_hook pm in
+  let pm = State.call_prg_hook first s_hook pm in
   let pm = List.fold_left progmap_remove pm l in
   pm, dref
 
@@ -1360,7 +1359,6 @@ let obligation_admitted_terminator ~pm {name; num; auto} uctx' dref =
       (Univ.UContext.instance uctx, uctx')
   in
   let obl = {obl with obl_body = Some (DefinedObl (cst, inst))} in
-  let () = if transparent then add_hint true prg cst in
   update_program_decl_on_defined ~pm prg obls num obl ~uctx:uctx' rem ~auto
 
 end
@@ -1998,10 +1996,11 @@ let compute_proof_using_for_admitted proof typ iproof =
       match Proof.partial_proof iproof with
       | pproof :: _ ->
         let env = Global.env () in
-        let ids_typ = Environ.global_vars_set env typ in
+        let sigma = (Proof.data iproof).Proof.sigma in
+        let ids_typ = Termops.global_vars_set env sigma typ in
         (* [pproof] is evar-normalized by [partial_proof]. We don't
            count variables appearing only in the type of evars. *)
-        let ids_def = Environ.global_vars_set env (EConstr.Unsafe.to_constr pproof) in
+        let ids_def = Termops.global_vars_set env sigma pproof in
         Some (Environ.really_needed env (Id.Set.union ids_typ ids_def))
       | [] -> None
 
@@ -2020,7 +2019,6 @@ let save_admitted ~pm ~proof =
     | [_, _, typ] -> typ
     | _ -> CErrors.anomaly ~label:"Lemmas.save_lemma_admitted" (Pp.str "more than one statement.")
   in
-  let typ = EConstr.Unsafe.to_constr typ in
   let iproof = get proof in
   let sec_vars = compute_proof_using_for_admitted proof typ iproof in
   let uctx = get_initial_euctx proof in
@@ -2210,8 +2208,6 @@ end
 
 let default_tactic = ref (Proofview.tclUNIT ())
 
-let evar_of_obligation o = Evd.make_evar (Global.named_context_val ()) (EConstr.of_constr o.obl_type)
-
 let subst_deps expand obls deps t =
   let osubst = Obls_.obl_substitution expand obls deps in
     (Vars.replace_vars (List.map (fun (n, (_, b)) -> n, b) osubst) t)
@@ -2220,8 +2216,6 @@ let subst_deps_obl obls obl =
   let t' = subst_deps true obls obl.obl_deps obl.obl_type in
   Obligation.set_type ~typ:t' obl
 
-open Evd
-
 let is_defined obls x = not (Option.is_empty obls.(x).obl_body)
 
 let deps_remaining obls deps =
@@ -2257,12 +2251,11 @@ let solve_by_tac prg obls i tac =
   let uctx = Internal.get_uctx prg in
   let uctx = UState.update_sigma_univs uctx (Global.universes ()) in
   let poly = Internal.get_poly prg in
-  let evi = evar_of_obligation obl in
   (* the status of [build_by_tactic] is dropped. *)
   try
     let env = Global.env () in
     let body, types, _univs, _, uctx =
-      build_by_tactic env ~uctx ~poly ~typ:evi.evar_concl tac in
+      build_by_tactic env ~uctx ~poly ~typ:(EConstr.of_constr obl.obl_type) tac in
     Inductiveops.control_only_guard env (Evd.from_ctx uctx) (EConstr.of_constr body);
     Some (body, types, uctx)
   with
@@ -2381,16 +2374,6 @@ let solve_all_obligations ~pm tac =
   State.fold pm ~init:pm ~f:(fun k v pm ->
       solve_prg_obligations ~pm v tac |> fst)
 
-let try_solve_obligation ~pm n prg tac =
-  let prg = get_unique_prog ~pm prg in
-  let {obls; remaining} = Internal.get_obligations prg in
-  let obls' = Array.copy obls in
-  match solve_obligation_by_tac prg obls' n tac with
-  | Some prg' ->
-    let pm, _ = update_obls ~pm prg' obls' (pred remaining) in
-    pm
-  | None -> pm
-
 let try_solve_obligations ~pm n tac =
   solve_obligations ~pm n tac |> fst
 
@@ -2460,6 +2443,7 @@ let msg_generating_obl name obls =
 
 let add_definition ~pm ~cinfo ~info ?obl_hook ?term ~uctx
     ?tactic ?(reduce = reduce) ?(opaque = false) obls =
+  let obl_hook = Option.map (fun h -> State.PrgHook h) obl_hook in
   let prg =
     ProgramDecl.make ~info ~cinfo ~body:term ~opaque ~uctx ~reduce ~ntns:[] ~deps:[] ~fixpoint_kind:None ?obl_hook obls
   in
@@ -2481,6 +2465,7 @@ let add_definition ~pm ~cinfo ~info ?obl_hook ?term ~uctx
 
 let add_mutual_definitions l ~pm ~info ?obl_hook ~uctx
     ?tactic ?(reduce = reduce) ?(opaque = false) ~ntns fixkind =
+  let obl_hook = Option.map (fun h -> State.PrgHook h) obl_hook in
   let deps = List.map (fun (ci,_,_) -> CInfo.get_name ci) l in
   let pm =
     List.fold_left
